import concurrent
import logging
import os
import re

from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from itertools import product

import numpy as np

from icenet.data.sic.mask import Masks
from icenet.data.sic.utils import SIC_HEMI_STR
from icenet.data.producers import Downloader
from icenet.data.utils import assign_lat_lon_coord_system, \
    gridcell_angles_from_dim_coords, \
    invert_gridcell_angles, \
    rotate_grid_vectors
from icenet.data.interfaces.utils import batch_requested_dates
from icenet.utils import run_command

import iris
import iris.exceptions
import pandas as pd

"""

"""


class ClimateDownloader(Downloader):
    """Climate downloader base class

    :param dates:
    :param delete_tempfiles:
    :param download:
    :param group_dates_by:
    :param max_threads:
    :param postprocess:
    :param pregrid_prefix:
    :param pressure_levels:
    :param var_name_idx:
    :param var_names:
    """

    def __init__(self, *args,
                 dates: object = (),
                 delete_tempfiles: bool = True,
                 download: bool = True,
                 group_dates_by: str = "year",
                 max_threads: int = 1,
                 postprocess: bool = True,
                 pregrid_prefix: str = "latlon_",
                 pressure_levels: object = (),
                 var_name_idx: int = -1,
                 var_names: object = (),
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._dates = list(dates)
        self._delete = delete_tempfiles
        self._download = download
        self._files_downloaded = []
        self._group_dates_by = group_dates_by
        self._masks = Masks(north=self.north, south=self.south)
        self._max_threads = max_threads
        self._postprocess = postprocess
        self._pregrid_prefix = pregrid_prefix
        self._pressure_levels = list(pressure_levels)
        self._sic_ease_cubes = dict()
        self._var_name_idx = var_name_idx
        self._var_names = list(var_names)

        assert len(self._var_names), "No variables requested"
        assert len(self._pressure_levels) == len(self._var_names), \
            "# of pressures must match # vars"

        self._download_method = None

        self._validate_config()

    def _validate_config(self):
        """

        """
        if self.hemisphere_str in os.path.split(self.base_path):
            raise RuntimeError("Don't include hemisphere string {} in "
                               "base path".format(self.hemisphere_str))

    def download(self):
        """

        """

        logging.info("Building request(s), downloading and daily averaging "
                     "from {} API".format(self.identifier.upper()))

        requests = list()

        for idx, var_name in enumerate(self.var_names):
            pressures = [None] if not self.pressure_levels[idx] else \
                self._pressure_levels[idx]

            dates_per_request = \
                batch_requested_dates(self._dates,
                                      attribute=self._group_dates_by)

            for var_prefix, pressure, req_date in \
                    product([var_name], pressures, dates_per_request):
                requests.append((var_prefix, pressure, req_date))

        with ThreadPoolExecutor(max_workers=
                                min(len(requests), self._max_threads)) \
                as executor:
            futures = []

            for var_prefix, pressure, req_date in requests:
                future = executor.submit(self._single_download,
                                         var_prefix,
                                         pressure,
                                         req_date)
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except Exception as e:
                    logging.exception("Thread failure: {}".format(e))

        logging.info("{} daily files downloaded".
                     format(len(self._files_downloaded)))

    def filter_dates_on_data(self,
                             var_prefix: str,
                             level: object,
                             req_dates: object):
        """

        :param var_prefix:
        :param level:
        :param req_dates:
        :return: req_dates(list), merge_files(list)
        """
        logging.warning("NOT IMPLEMENTED YET, WE'LL JUST DOWNLOAD ANYWAY")
        # TODO: check existing yearly file for req_dates if already in place
        return req_dates, []

    # FIXME: this can be migrated for Dev#45 and we register in the subclass
    #  the potential implementations (e.g. CDS has toolbox and API, CMEMS has
    #  up to four implementations)
    def _single_download(self,
                         var_prefix: str,
                         level: object,
                         req_dates: object):
        """Implements a single download from CMEMS

        :param var_prefix: the icenet variable name
        :param level: the height to download
        :param req_dates: the request date
        """

        logging.info("Processing single download for {} @ {} with {} dates".
                     format(var_prefix, level, len(req_dates)))
        var = var_prefix if not level else \
            "{}{}".format(var_prefix, level)
        var_folder = self.get_data_var_folder(var)

        latlon_path, regridded_name = \
            self.get_req_filenames(var_folder, req_dates[0])

        req_dates, merge_files = \
            self.filter_dates_on_data(var_prefix, level, req_dates)

        if self.download and not os.path.exists(latlon_path):
            self.download_method(var,
                                 level,
                                 req_dates,
                                 latlon_path)

            logging.info("Downloaded to {}".format(latlon_path))
        else:
            logging.info("Skipping actual download")

        if self._postprocess:
            self.postprocess(var, latlon_path)

        if not os.path.exists(regridded_name):
            self._files_downloaded.append(latlon_path)

    def postprocess(self, var, download_path):
        logging.debug("No postprocessing in place for {}: {}".
                      format(var, download_path))

    def save_temporal_files(self, var, da,
                            date_format=None,
                            freq=None):
        """

        :param var:
        :param da:
        :param date_format:
        :param freq:
        """
        var_folder = self.get_data_var_folder(var)
        group_by = "time.{}".format(self._group_dates_by) if not freq else freq

        for dt, dt_da in da.groupby(group_by):
            req_date = pd.to_datetime(dt_da.time.values[0])
            latlon_path, regridded_name = \
                self.get_req_filenames(var_folder,
                                       req_date,
                                       date_format=date_format)

            logging.info("Retrieving and saving {}".format(latlon_path))
            dt_da.compute()
            dt_da.to_netcdf(latlon_path)

            if not os.path.exists(regridded_name):
                self._files_downloaded.append(latlon_path)

    @property
    def sic_ease_cube(self):
        """

        :return sic_cube:
        """
        if self._hemisphere not in self._sic_ease_cubes:
            sic_day_fname = 'ice_conc_{}_ease2-250_cdr-v2p0_197901021200.nc'. \
                format(SIC_HEMI_STR[self.hemisphere_str[0]])
            sic_day_path = os.path.join(self.get_data_var_folder("siconca"),
                                        sic_day_fname)
            if not os.path.exists(sic_day_path):
                logging.info("Downloading single daily SIC netCDF file for "
                             "regridding ERA5 data to EASE grid...")

                retrieve_sic_day_cmd = 'wget -m -nH --cut-dirs=6 -P {} ' \
                                       'ftp://osisaf.met.no/reprocessed/ice/' \
                                       'conc/v2p0/1979/01/{}'.\
                    format(self.get_data_var_folder("siconca"), sic_day_fname)

                run_command(retrieve_sic_day_cmd)

            # Load a single SIC map to obtain the EASE grid for
            # regridding ERA data
            self._sic_ease_cubes[self._hemisphere] = \
                iris.load_cube(sic_day_path, 'sea_ice_area_fraction')

            # Convert EASE coord units to metres for regridding
            self._sic_ease_cubes[self._hemisphere].coord(
                'projection_x_coordinate').convert_units('meters')
            self._sic_ease_cubes[self._hemisphere].coord(
                'projection_y_coordinate').convert_units('meters')
        return self._sic_ease_cubes[self._hemisphere]

    def regrid(self,
               files: object = None):
        """

        :param files:
        """
        filelist = self._files_downloaded if not files else files
        batches = [filelist[b:b + 1000] for b in range(0, len(filelist), 1000)]

        max_workers = min(len(batches), self._max_threads)

        if max_workers > 0:
            with ThreadPoolExecutor(max_workers=max_workers) \
                    as executor:
                futures = []

                for files in batches:
                    future = executor.submit(self._batch_regrid, files)
                    futures.append(future)

                for future in concurrent.futures.as_completed(futures):
                    try:
                        future.result()
                    except Exception as e:
                        logging.exception("Thread failure: {}".format(e))
        else:
            logging.info("No regrid batches to processing, moving on...")

    def _batch_regrid(self,
                      files: object):
        """

        :param files:
        """
        for datafile in files:
            (datafile_path, datafile_name) = os.path.split(datafile)
            # TODO: mmmm, need to keep consistent with get_daily_filenames
            new_datafile = os.path.join(datafile_path,
                                        re.sub(
                                            r'^{}'.format(self.pregrid_prefix),
                                            '', datafile_name))

            if os.path.exists(new_datafile):
                logging.debug("Skipping {} as {} already exists".
                    format(datafile, os.path.basename(new_datafile)))
                continue

            logging.debug("Regridding {}".format(datafile))

            try:
                cube = iris.load_cube(datafile)
                cube = self.convert_cube(cube)

                cube_ease = cube.regrid(
                    self.sic_ease_cube, iris.analysis.Linear())

            except iris.exceptions.CoordinateNotFoundError:
                logging.warning("{} has no coordinates...".
                                format(datafile_name))
                if self.delete:
                    logging.debug("Deleting failed file {}...".
                                  format(datafile_name))
                    os.unlink(datafile)
                continue

            self.additional_regrid_processing(datafile, cube_ease)

            # TODO: filename chain can be handled better for sharing between
            #  methods
            logging.info("Saving regridded data to {}... ".format(new_datafile))
            iris.save(cube_ease, new_datafile, fill_value=np.nan)

            if self.delete:
                logging.info("Removing {}".format(datafile))
                os.remove(datafile)

    def convert_cube(self, cube: object):
        """Converts Iris cube to be fit for regrid

        :param cube: the cube requiring alteration
        :return cube: the altered cube
        """

        cube = assign_lat_lon_coord_system(cube)
        return cube

    @abstractmethod
    def additional_regrid_processing(self,
                                     datafile: str,
                                     cube_ease: object):
        """

        :param datafile:
        :param cube_ease:
        """
        pass

    def rotate_wind_data(self,
                         apply_to: object = ("uas", "vas"),
                         manual_files: object = None):
        """

        :param apply_to:
        :param manual_files:
        """
        assert len(apply_to) == 2, "Too many wind variables supplied: {}, " \
                                   "there should only be two.".\
            format(", ".join(apply_to))

        angles = gridcell_angles_from_dim_coords(self.sic_ease_cube)
        invert_gridcell_angles(angles)

        logging.info("Rotating wind data in {}".format(
            " ".join([self.get_data_var_folder(v) for v in apply_to])))

        wind_files = {}

        for var in apply_to:
            source = self.get_data_var_folder(var)

            file_source = self._files_downloaded \
                if not manual_files else manual_files

            latlon_files = [df for df in file_source if source in df]
            wind_files[var] = sorted([
                re.sub(r'{}'.format(self.pregrid_prefix), '', df)
                for df in latlon_files
                if os.path.dirname(df).split(os.sep)
                [self._var_name_idx] == var],
                key=lambda x: int(re.search(r'^(?:\w+_)?(\d+).nc',
                                  os.path.basename(x)).group(1))
            )
            logging.info("{} files for {}".format(len(wind_files[var]), var))

        # NOTE: we're relying on apply_to having equal datasets
        assert len(wind_files[apply_to[0]]) == len(wind_files[apply_to[1]]), \
            "The wind file datasets are unequal in length"

        # validation
        for idx, wind_file_0 in enumerate(wind_files[apply_to[0]]):
            wind_file_1 = wind_files[apply_to[1]][idx]

            wd0 = re.sub(r'^{}_'.format(apply_to[0]), '',
                         os.path.basename(wind_file_0))

            if not wind_file_1.endswith(wd0):
                logging.error("Wind file array is not valid:".format(
                    zip(wind_files)))
                raise RuntimeError("{} is not at the end of {}, something is "
                                   "wrong".format(wd0, wind_file_1))

        for idx, wind_file_0 in enumerate(wind_files[apply_to[0]]):
            wind_file_1 = wind_files[apply_to[1]][idx]

            logging.info("Rotating {} and {}".format(wind_file_0, wind_file_1))

            wind_cubes = dict()
            wind_cubes_r = dict()

            wind_cubes[apply_to[0]] = iris.load_cube(wind_file_0)
            wind_cubes[apply_to[1]] = iris.load_cube(wind_file_1)

            try:
                wind_cubes_r[apply_to[0]], wind_cubes_r[apply_to[1]] = \
                    rotate_grid_vectors(
                        wind_cubes[apply_to[0]],
                        wind_cubes[apply_to[1]],
                        angles,
                    )
            except iris.exceptions.CoordinateNotFoundError:
                logging.exception("Failure to rotate due to coordinate issues. "
                                  "moving onto next file")
                continue

            # Original implementation is in danger of lost updates
            # due to potential lazy loading
            for i, name in enumerate([wind_file_0, wind_file_1]):
                # NOTE: implementation with tempfile caused problems on NFS
                # mounted filesystem, so avoiding in place of letting iris do it
                temp_name = os.path.join(os.path.split(name)[0],
                                         "temp.{}".format(
                                             os.path.basename(name)))
                logging.debug("Writing {}".format(temp_name))

                iris.save(wind_cubes_r[apply_to[i]], temp_name)
                os.replace(temp_name, name)

    def get_req_filenames(self,
                          var_folder: str,
                          req_date: object,
                          date_format: str = None):
        """

        :param var_folder:
        :param req_date:
        :param date_format:
        :return:
        """

        filename_date = getattr(req_date, self._group_dates_by) \
            if not date_format else req_date.strftime(date_format)

        latlon_path = os.path.join(
            var_folder, "{}{}.nc".format(self.pregrid_prefix, filename_date))
        regridded_name = os.path.join(
            var_folder, "{}.nc".format(filename_date))

        logging.debug("Got {} filenames: {} and {}".format(
            self._group_dates_by, latlon_path, regridded_name
        ))

        return latlon_path, regridded_name

    @property
    def dates(self):
        return self._dates

    @property
    def delete(self):
        return self._delete

    @property
    def download_method(self) -> callable:
        if not self._download_method:
            raise RuntimeError("Downloader has no method set, "
                               "implementation error")
        return self._download_method

    @download_method.setter
    def download_method(self, method: callable):
        self._download_method = method

    @property
    def group_dates_by(self):
        return self._group_dates_by

    @property
    def pregrid_prefix(self):
        return self._pregrid_prefix

    @property
    def pressure_levels(self):
        return self._pressure_levels

    @property
    def var_names(self):
        return self._var_names




import numpy as np
import tensorflow as tf


class IceNetPreTrainingEvaluator(tf.keras.callbacks.Callback):
    """
    Custom tf.keras callback to update the `logs` dict used by all other callbacks
    with the validation set metrics. The callback is executed every
    `validation_frequency` batches.

    This can be used in conjuction with the BatchwiseModelCheckpoint callback to
    perform a model checkpoint based on validation data every N batches - ensure
    the `save_frequency` input to BatchwiseModelCheckpoint is also set to
    `validation_frequency`.

    Also ensure that the callbacks list past to Model.fit() contains this
    callback before any other callbacks that need the validation metrics.

    Also use Weights and Biases to log the training and validation metrics.

    TODO: not sure this is really necessary/stable, review

    :param validation_frequency:
    :param val_dataloader:
    :param sample_at_zero:
    """

    def __init__(self,
                 *args,
                 validation_frequency: int,
                 val_dataloader: object,
                 sample_at_zero: bool = False,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.validation_frequency = validation_frequency
        self.val_dataloader = val_dataloader
        self.sample_at_zero = sample_at_zero

    def on_train_batch_end(self,
                           batch: object,
                           logs: object = None):
        """

        :param batch:
        :param logs:
        """
        if (batch == 0 and self.sample_at_zero) \
                or (batch + 1) % self.validation_frequency == 0:
            val_logs = self.model.evaluate(self.val_dataloader,
                                           verbose=0,
                                           return_dict=True)
            val_logs = {'val_' + name: val for name, val in val_logs.items()}
            logs.update(val_logs)
            [tf.print('\n' + k + ' {:.2f}'.format(v)) for k, v in logs.items()]
            tf.print('\n')


class BatchwiseModelCheckpoint(tf.keras.callbacks.Callback):
    """

    :param save_frequency:
    :param model_path:
    :param mode:
    :param monitor:
    :param prev_best:
    :param sample_at_zero:
    """

    def __init__(self,
                 save_frequency: object,
                 model_path: object,
                 mode: object,
                 monitor: object,
                 prev_best: object = None,
                 sample_at_zero: object = False):
        super().__init__()
        self.save_frequency = save_frequency
        self.model_path = model_path
        self.mode = mode
        self.monitor = monitor
        self.sample_at_zero = sample_at_zero

        if prev_best is not None:
            self.best = prev_best

        else:
            if self.mode == 'max':
                self.best = -np.Inf
            elif self.mode == 'min':
                self.best = np.Inf

    def on_train_batch_end(self,
                           batch: object,
                           logs: object = None):
        """

        :param batch:
        :param logs:
        """
        if (batch == 0 and self.sample_at_zero) \
                or (batch + 1) % self.save_frequency == 0:
            if self.mode == 'max' and logs[self.monitor] > self.best:
                save = True

            elif self.mode == 'min' and logs[self.monitor] < self.best:
                save = True

            else:
                save = False

            if save:
                tf.print('\n{} improved from {:.3f} to {:.3f}. '
                         'Saving model to {}.\n'.
                         format(self.monitor,
                                self.best,
                                logs[self.monitor],
                                self.model_path))

                self.best = logs[self.monitor]

                self.model.save(self.model_path, overwrite=True)
            else:
                tf.print('\n{}={:.3f} did not improve from {:.3f}\n'.format(
                    self.monitor, logs[self.monitor], self.best))



def plot_linear_trend(loader, var_name, date, trend_leadtime):
    """Plot linear trend against source data

    This is a method to plot linear trends against the corresponding variable
    data from a given preprocessing run.

    :return:
    """



import glob
import logging
import os

import numpy as np
import tensorflow as tf


def get_decoder(shape: object,
                channels: object,
                forecasts: object,
                num_vars: int = 1,
                dtype: str = "float32") -> object:
    """

    :param shape:
    :param channels:
    :param forecasts:
    :param num_vars:
    :param dtype:
    :return:
    """
    xf = tf.io.FixedLenFeature(
        [*shape, channels], getattr(tf, dtype))
    yf = tf.io.FixedLenFeature(
        [*shape, forecasts, num_vars], getattr(tf, dtype))
    sf = tf.io.FixedLenFeature(
        [*shape, forecasts, num_vars], getattr(tf, dtype))

    @tf.function
    def decode_item(proto):
        features = {
            "x": xf,
            "y": yf,
            "sample_weights": sf,
        }

        item = tf.io.parse_example(proto, features)
        return item['x'], item['y'], item['sample_weights']

    return decode_item


# TODO: define a decent interface and sort the inheritance architecture out, as
#  this will facilitate the new datasets in #35
class SplittingMixin:
    """

    """

    _batch_size: int
    _dtype: object
    _num_channels: int
    _n_forecast_days: int
    _shape: int
    _shuffling: bool

    train_fns = []
    test_fns = []
    val_fns = []

    def add_records(self, base_path: str, hemi: str):
        """

        :param base_path:
        :param hemi:
        """
        train_path = os.path.join(base_path, hemi, "train")
        val_path = os.path.join(base_path, hemi, "val")
        test_path = os.path.join(base_path, hemi, "test")

        logging.info("Training dataset path: {}".format(train_path))
        self.train_fns += glob.glob("{}/*.tfrecord".format(train_path))
        logging.info("Validation dataset path: {}".format(val_path))
        self.val_fns += glob.glob("{}/*.tfrecord".format(val_path))
        logging.info("Test dataset path: {}".format(test_path))
        self.test_fns += glob.glob("{}/*.tfrecord".format(test_path))

    def get_split_datasets(self, ratio: object = None):
        """

        :param ratio:
        :return:
        """
        if not (len(self.train_fns) + len(self.val_fns) + len(self.test_fns)):
            raise RuntimeError("No files have been found, abandoning...")

        logging.info("Datasets: {} train, {} val and {} test filenames".format(
            len(self.train_fns), len(self.val_fns), len(self.test_fns)))

        if ratio:
            if ratio > 1.0:
                raise RuntimeError("Ratio cannot be more than 1")

            logging.info("Reducing datasets to {} of total files".format(ratio))
            train_idx, val_idx, test_idx = \
                int(len(self.train_fns) * ratio), \
                int(len(self.val_fns) * ratio), \
                int(len(self.test_fns) * ratio)

            if train_idx > 0:
                self.train_fns = self.train_fns[:train_idx]
            if val_idx > 0:
                self.val_fns = self.val_fns[:val_idx]
            if test_idx > 0:
                self.test_fns = self.test_fns[:test_idx]

            logging.info("Reduced: {} train, {} val and {} test filenames".format(
                len(self.train_fns), len(self.val_fns), len(self.test_fns)))

        train_ds, val_ds, test_ds = \
            tf.data.TFRecordDataset(self.train_fns,
                                    num_parallel_reads=self.batch_size), \
            tf.data.TFRecordDataset(self.val_fns,
                                    num_parallel_reads=self.batch_size), \
            tf.data.TFRecordDataset(self.test_fns,
                                    num_parallel_reads=self.batch_size),

        # TODO: Comparison/profiling runs
        # TODO: parallel for batch size while that's small
        # TODO: obj.decode_item might not work here - figure out runtime
        #  implementation based on wrapped function call that can be serialised
        decoder = get_decoder(self.shape,
                              self.num_channels,
                              self.n_forecast_days,
                              dtype=self.dtype.__name__)

        if self.shuffling:
            logging.info("Training dataset(s) marked to be shuffled")
            # FIXME: this is not a good calculation, but we don't have access
            #  in the mixin to the configuration that generated the dataset #57
            train_ds = train_ds.shuffle(
                min(int(len(self.train_fns) * self.batch_size), 366))

        train_ds = train_ds.\
            map(decoder, num_parallel_calls=self.batch_size).\
            batch(self.batch_size)

        val_ds = val_ds.\
            map(decoder, num_parallel_calls=self.batch_size).\
            batch(self.batch_size)

        test_ds = test_ds.\
            map(decoder, num_parallel_calls=self.batch_size).\
            batch(self.batch_size)

        return train_ds.prefetch(tf.data.AUTOTUNE), \
            val_ds.prefetch(tf.data.AUTOTUNE), \
            test_ds.prefetch(tf.data.AUTOTUNE)

    def check_dataset(self,
                      split: str = "train"):
        logging.debug("Checking dataset {}".format(split))

        decoder = get_decoder(self.shape,
                              self.num_channels,
                              self.n_forecast_days,
                              dtype=self.dtype.__name__)

        for df in getattr(self, "{}_fns".format(split)):
            logging.debug("Getting records from {}".format(df))
            try:
                raw_dataset = tf.data.TFRecordDataset([df])
                raw_dataset = raw_dataset.map(decoder)

                for i, (x, y, sw) in enumerate(raw_dataset):
                    x = x.numpy()
                    y = y.numpy()
                    sw = sw.numpy()

                    logging.debug("Got record {} with x {} y {} sw {}".
                                  format(i,
                                         x.shape,
                                         y.shape,
                                         sw.shape))

                    input_nans = np.isnan(x).sum()
                    output_nans = np.isnan(y[sw > 0.]).sum()

                    if input_nans > 0:
                        logging.warning("Input NaNs detected in {}:{}".
                                        format(df, i))

                    if output_nans > 0:
                        logging.warning("Output NaNs detected in {}:{}, not "
                                        "accounted for by sample weighting".
                                        format(df, i))
            except tf.errors.DataLossError as e:
                logging.warning("{}: data loss error {}".format(df, e.message))
            except tf.errors.OpError as e:
                logging.warning("{}: tensorflow error {}".format(df, e.message))
            # We don't except any non-tensorflow errors to prevent progression

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def dtype(self):
        return self._dtype

    @property
    def n_forecast_days(self):
        return self._n_forecast_days

    @property
    def num_channels(self):
        return self._num_channels

    @property
    def shape(self):
        return self._shape

    @property
    def shuffling(self):
        return self._shuffling



import argparse
import logging
import os
import shutil

import xarray as xr

from icenet.process.utils import date_arg, destination_filename
from icenet.utils import setup_logging


@setup_logging
def upload_parse_args():
    """

    :return:
    """
    a = argparse.ArgumentParser()

    a.add_argument("filename")
    a.add_argument("destination")
    a.add_argument("date", default=None, type=date_arg, nargs="?")

    a.add_argument("-v", "--verbose", default=False, action="store_true")

    return a.parse_args()


def upload():
    """

    """
    args = upload_parse_args()
    logging.info("Local upload facility")

    if not os.path.isdir(args.destination):
        raise RuntimeError("Destination {} does not exist".
                           format(args.destination))

    if args.date:
        ds = xr.open_dataset(args.filename)
        ds = ds.sel(time=slice(args.date, args.date))

        if len(ds.time) < 1:
            raise ValueError("No elements in {} for {}".format(
                args.filename, args.date
            ))

        filename = destination_filename(args.destination,
                                        args.filename,
                                        args.date)
        ds.to_netcdf(filename)
        ds.close()

        logging.info("Saved to {}".format(filename))
    else:
        newname = shutil.copy(args.filename, args.destination)
        logging.info("Copied to {}".format(newname))



from icenet.data.cli import process_args, process_date_args
from icenet.data.process import IceNetPreProcessor

"""

"""


class IceNetERA5PreProcessor(IceNetPreProcessor):
    """

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         file_filters=["latlon_"],
                         identifier="era5",
                         **kwargs)


def main():
    args = process_args()
    dates = process_date_args(args)

    era5 = IceNetERA5PreProcessor(
        args.abs,
        args.anom,
        args.name,
        dates["train"],
        dates["val"],
        dates["test"],
        linear_trends=args.trends,
        linear_trend_days=args.trend_lead,
        north=args.hemisphere == "north",
        ref_procdir=args.ref,
        south=args.hemisphere == "south",
        update_key=args.update_key,
    )
    era5.init_source_data(
        lag_days=args.lag,
    )
    era5.process()



import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Conv2D, BatchNormalization, UpSampling2D, \
    concatenate, MaxPooling2D, Input
from tensorflow.keras.optimizers import Adam

"""
Defines the Python-based sea ice forecasting models, such as the IceNet architecture
and the linear trend extrapolation model.
"""


@tf.keras.utils.register_keras_serializable()
class TemperatureScale(tf.keras.layers.Layer):
    """Temperature scaling layer

    Implements the temperature scaling layer for probability calibration,
    as introduced in Guo 2017 (http://proceedings.mlr.press/v70/guo17a.html).
    """
    def __init__(self, **kwargs):
        super(TemperatureScale, self).__init__(**kwargs)
        self.temp = tf.Variable(initial_value=1.0, trainable=False,
                                dtype=tf.float32, name='temp')

    def call(self, inputs: object, **kwargs):
        """ Divide the input logits by the T value.

        :param **kwargs:
        :param inputs:
        :return:
        """
        return tf.divide(inputs, self.temp)

    def get_config(self):
        """ For saving and loading networks with this custom layer.

        :return:
        """
        return {'temp': self.temp.numpy()}


### Network architectures:
# --------------------------------------------------------------------

def unet_batchnorm(input_shape: object,
                   loss: object,
                   metrics: object,
                   learning_rate: float = 1e-4,
                   filter_size: float = 3,
                   n_filters_factor: float = 1,
                   n_forecast_days: int = 1) -> object:
    """

    :param input_shape:
    :param loss:
    :param metrics:
    :param learning_rate:
    :param filter_size:
    :param n_filters_factor:
    :param n_forecast_days:
    :return:
    """
    inputs = Input(shape=input_shape)

    conv1 = Conv2D(np.int(64*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(inputs)
    conv1 = Conv2D(np.int(64*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv1)
    bn1 = BatchNormalization(axis=-1)(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(bn1)

    conv2 = Conv2D(np.int(128*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(pool1)
    conv2 = Conv2D(np.int(128*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv2)
    bn2 = BatchNormalization(axis=-1)(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(bn2)

    conv3 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(pool2)
    conv3 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv3)
    bn3 = BatchNormalization(axis=-1)(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(bn3)

    conv4 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(pool3)
    conv4 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv4)
    bn4 = BatchNormalization(axis=-1)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(bn4)

    conv5 = Conv2D(np.int(512*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(pool4)
    conv5 = Conv2D(np.int(512*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv5)
    bn5 = BatchNormalization(axis=-1)(conv5)

    up6 = Conv2D(np.int(256*n_filters_factor), 2, activation='relu',
                 padding='same', kernel_initializer='he_normal')(
        UpSampling2D(size=(2, 2), interpolation='nearest')(bn5))

    merge6 = concatenate([bn4, up6], axis=3)
    conv6 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(merge6)
    conv6 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv6)
    bn6 = BatchNormalization(axis=-1)(conv6)

    up7 = Conv2D(np.int(256*n_filters_factor), 2, activation='relu',
                 padding='same', kernel_initializer='he_normal')(
        UpSampling2D(size=(2, 2), interpolation='nearest')(bn6))
    merge7 = concatenate([bn3,up7], axis=3)
    conv7 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(merge7)
    conv7 = Conv2D(np.int(256*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv7)
    bn7 = BatchNormalization(axis=-1)(conv7)

    up8 = Conv2D(np.int(128*n_filters_factor), 2, activation='relu',
                 padding='same', kernel_initializer='he_normal')(
        UpSampling2D(size=(2, 2), interpolation='nearest')(bn7))
    merge8 = concatenate([bn2,up8], axis=3)
    conv8 = Conv2D(np.int(128*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(merge8)
    conv8 = Conv2D(np.int(128*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv8)
    bn8 = BatchNormalization(axis=-1)(conv8)

    up9 = Conv2D(np.int(64*n_filters_factor), 2, activation='relu',
                 padding='same', kernel_initializer='he_normal')(
        UpSampling2D(size=(2, 2), interpolation='nearest')(bn8))

    merge9 = concatenate([conv1, up9], axis=3)

    conv9 = Conv2D(np.int(64*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(merge9)
    conv9 = Conv2D(np.int(64*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv9)
    conv9 = Conv2D(np.int(64*n_filters_factor), filter_size, activation='relu',
                   padding='same', kernel_initializer='he_normal')(conv9)

    final_layer = Conv2D(n_forecast_days,
                         kernel_size=1, activation='sigmoid')(conv9)

    # Keras graph mode needs y_pred and y_true to have the same shape, so we
    #   we must pad an extra dimension onto the model output to train with
    #   an extra sample weight dimension in y_true.
    # final_layer = tf.expand_dims(final_layer, axis=-1)

    model = Model(inputs, final_layer)

    model.compile(
        optimizer=Adam(learning_rate=learning_rate),
        loss=loss,
        weighted_metrics=metrics)

    return model


def linear_trend_forecast(usable_selector: object,
                          forecast_date: object,
                          da: object,
                          mask: object,
                          missing_dates: object = (),
                          shape: object = (432, 432)) -> object:
    """

    :param usable_selector:
    :param forecast_date:
    :param da:
    :param mask:
    :param missing_dates:
    :param shape:
    :return:
    """
    usable_data = usable_selector(da, forecast_date, missing_dates)

    if len(usable_data.time) < 1:
        return np.full(shape, np.nan)

    x = np.arange(len(usable_data.time))
    y = usable_data.data.reshape(len(usable_data.time), -1)

    src = np.c_[x, np.ones_like(x)]
    r = np.linalg.lstsq(src, y, rcond=None)[0]
    output_map = np.matmul(
        np.array([len(usable_data.time), 1]), r).reshape(*shape)
    output_map[mask] = 0.
    output_map[output_map < 0] = 0.
    output_map[output_map > 1] = 1.

    return output_map



import argparse
import datetime as dt
import glob
import logging
import os
import re

import iris
import numpy as np
import pandas as pd
import xarray as xr

from icenet import __version__ as icenet_version
from icenet.data.dataset import IceNetDataSet
from icenet.data.sic.mask import Masks
from icenet.utils import run_command, setup_logging


def get_refsic(north: bool = True, south: bool = False) -> object:
    """

    :param north:
    :param south:
    :return:
    """
    assert north or south, "Select one hemisphere at least..."

    str = "nh" if north else "sh"

    sic_day_fname = 'ice_conc_{}_ease2-250_cdr-v2p0_197901021200.nc'.format(str)
    sic_day_path = os.path.join(".", "_sicfile")

    if not os.path.exists(os.path.join(sic_day_path, sic_day_fname)):
        logging.info("Downloading single daily SIC netCDF file for "
                     "regridding ERA5 data to EASE grid...")

        retrieve_sic_day_cmd = 'wget -m -nH --cut-dirs=6 -P {} ' \
                               'ftp://osisaf.met.no/reprocessed/ice/' \
                               'conc/v2p0/1979/01/{}'. \
            format(sic_day_path, sic_day_fname)

        run_command(retrieve_sic_day_cmd)

    return os.path.join(sic_day_path, sic_day_fname)


def get_refcube(north: bool = True, south: bool = False) -> object:
    """

    :param north:
    :param south:
    :return:
    """
    assert north or south, "Select one hemisphere at least..."

    path = get_refsic(north, south)

    cube = iris.load_cube(path, 'sea_ice_area_fraction')
    return cube


def get_prediction_data(root: object,
                        name: object,
                        date: object) -> object:
    """

    :param root:
    :param name:
    :param date:
    :return:
    """
    logging.info("Post-processing {}".format(date))

    glob_str = os.path.join(root,
                            "results",
                            "predict",
                            name,
                            "*",
                            date.strftime("%Y_%m_%d.npy"))

    np_files = glob.glob(glob_str)
    if not len(np_files):
        logging.warning("No files found")
        return None

    data = [np.load(f) for f in np_files]
    data = np.array(data)

    logging.debug("Data read from disk: {} from: {}".format(data.shape, np_files))

    return np.stack(
        [data.mean(axis=0), data.std(axis=0)],
        axis=-1).squeeze()


def date_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    date_match = re.search(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)
    return dt.date(*[int(s) for s in date_match.groups()])


@setup_logging
def get_args():
    """

    :return:
    """
    ap = argparse.ArgumentParser()
    ap.add_argument("name")
    ap.add_argument("dataset")
    ap.add_argument("datefile", type=argparse.FileType("r"))

    ap.add_argument("-m", "--mask", default=False, action="store_true")
    ap.add_argument("-o", "--output-dir", default=".")
    ap.add_argument("-r", "--root", type=str, default=".")

    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    return ap.parse_args()


def create_cf_output():
    """

    """
    args = get_args()

    dataset_config = \
        os.path.join(args.root, "dataset_config.{}.json".format(args.dataset))
    ds = IceNetDataSet(dataset_config)

    ref_sic = xr.open_dataset(get_refsic(ds.north, ds.south))
    ref_cube = get_refcube(ds.north, ds.south)

    dates = [dt.date(*[int(v) for v in s.split("-")])
             for s in args.datefile.read().split()]
    args.datefile.close()

    arr = np.array(
        [get_prediction_data(args.root, args.name, date)
         for date in dates])

    logging.info("Dataset arr shape: {}".format(arr.shape))

    sic_mean = arr[..., 0]
    sic_stddev = arr[..., 1]

    if args.mask:
        mask_gen = Masks(north=ds.north, south=ds.south)

        logging.info("Land masking the forecast output")
        land_mask = mask_gen.get_land_mask()
        mask = land_mask[np.newaxis, ..., np.newaxis]
        mask = np.repeat(mask, sic_mean.shape[-1], axis=-1)
        mask = np.repeat(mask, sic_mean.shape[0], axis=0)

        sic_mean[mask] = 0
        sic_stddev[mask] = 0

        logging.info("Applying active grid cell masks")

        for idx, forecast_date in enumerate(dates):
            for lead_idx in np.arange(0, arr.shape[3], 1):
                lead_dt = forecast_date + dt.timedelta(days=int(lead_idx) + 1)
                logging.debug("Active grid cell mask start {} forecast date {}".
                              format(forecast_date, lead_dt))

                grid_cell_mask = mask_gen.get_active_cell_mask(lead_dt.month)
                sic_mean[idx, ~grid_cell_mask, lead_idx] = 0
                sic_stddev[idx, ~grid_cell_mask, lead_idx] = 0

    xarr = xr.Dataset(
        data_vars=dict(
            Lambert_Azimuthal_Grid=ref_sic.Lambert_Azimuthal_Grid,
            sic_mean=(["time", "yc", "xc", "leadtime"], sic_mean),
            sic_stddev=(["time", "yc", "xc", "leadtime"], sic_stddev),
        ),
        coords=dict(
            time=[pd.Timestamp(d) for d in dates],
            leadtime=np.arange(1, arr.shape[3] + 1, 1),
            xc=ref_cube.coord("projection_x_coordinate").points,
            yc=ref_cube.coord("projection_y_coordinate").points,
            lat=(("yc", "xc"), ref_cube.coord("latitude").points),
            lon=(("yc", "xc"), ref_cube.coord("longitude").points),
        ),
        # https://wiki.esipfed.org/Attribute_Convention_for_Data_Discovery_1-3
        attrs=dict(
            Conventions="CF-1.6 ACDD-1.3",
            comments="",
            creator_email="jambyr@bas.ac.uk",
            creator_institution="British Antarctic Survey",
            creator_name="James Byrne",
            creator_url="www.bas.ac.uk",
            date_created=dt.datetime.now().strftime("%Y-%m-%d"),
            # Issue#18: Overcoming OSI-SAF EPSG ref issue
            geospatial_bounds_crs="EPSG:6931" if ds.north else "EPSG:6932",
            geospatial_lat_min=ref_cube.attributes["geospatial_lat_min"],
            geospatial_lat_max=ref_cube.attributes["geospatial_lat_max"],
            geospatial_lon_min=ref_cube.attributes["geospatial_lon_min"],
            geospatial_lon_max=ref_cube.attributes["geospatial_lon_max"],
            geospatial_vertical_min=0.0,
            geospatial_vertical_max=0.0,
            history="{} - creation".format(dt.datetime.now()),
            id="IceNet.TBC".format(icenet_version),
            institution="British Antarctic Survey",
            keywords="""'Earth Science > Cryosphere > Sea Ice > Sea Ice Concentration
            Earth Science > Oceans > Sea Ice > Sea Ice Concentration
            Earth Science > Climate Indicators > Cryospheric Indicators > Sea Ice
            Geographic Region > {} Hemisphere""".format(
                "Northern" if ds.north else "Southern"
            ),
            # TODO: check we're valid
            keywords_vocabulary="GCMD Science Keywords",
            # TODO: Double check this is good with PDC
            license="Open Government Licece (OGL) V3",
            naming_authority="uk.ac.bas",
            platform="BAS HPC",
            #program="",
            #processing_level="",
            product_version=icenet_version,
            project="IceNet",
            publisher_email="",
            publisher_institution="British Antarctic Survey",
            #publisher_name="Polar Data Center",
            publisher_url="",
            source="""
            IceNet model generation at v{}
            """.format(icenet_version),
            spatial_resolution=ref_cube.attributes["spatial_resolution"],
            # Values for any standard_name attribute must come from the CF
            # Standard Names vocabulary for the data file or product to
            #  comply with CF
            standard_name_vocabulary="CF Standard Name Table v27",
            summary="""
            This is an output of sea ice concentration predictions from the 
            IceNet UNet run in an ensemble, with postprocessing to determine 
            the mean and standard deviation across the runs.
            """,
            # Use ISO 8601:2004 duration format, preferably the extended format
            # as recommended in the Attribute Content Guidance section.
            time_coverage_start="",
            time_coverage_end="",
            time_coverage_duration="P1D",
            time_coverage_resolution="P1D",
            title="Sea Ice Concentration Prediction",
        )
    )

    xarr.time.attrs = dict(
        long_name=ref_cube.coord("time").long_name,
        standard_name=ref_cube.coord("time").standard_name,
        axis="T",
        # TODO: https://github.com/SciTools/cf-units for units methods
        # units=Unit('seconds since 1978-01-01 00:00:00', calendar='gregorian')
        # bounds=array([[31622400., 31708800.]])
    )
    xarr.yc.attrs = dict(
        long_name=ref_cube.coord("projection_y_coordinate").long_name,
        standard_name=ref_cube.coord("projection_y_coordinate").standard_name,
        units=ref_cube.coord("projection_y_coordinate").units.name,
        axis="Y",
        # TODO: iris.coord_systems.LambertAzimuthalEqualArea
    )
    xarr.xc.attrs = dict(
        long_name=ref_cube.coord("projection_x_coordinate").long_name,
        standard_name=ref_cube.coord("projection_x_coordinate").standard_name,
        units=ref_cube.coord("projection_x_coordinate").units.name,
        axis="X",
    )
    xarr.leadtime.attrs = dict(
        long_name="leadtime of forecast in relation to reference time",
        short_name="leadtime",
        #units="1",
    )

    xarr.lat.attrs = dict(
        long_name=ref_cube.coord("latitude").long_name,
        standard_name=ref_cube.coord("latitude").standard_name,
        units=ref_cube.coord("latitude").units.name,
    )
    xarr.lon.attrs = dict(
        long_name=ref_cube.coord("longitude").long_name,
        standard_name=ref_cube.coord("longitude").standard_name,
        units=ref_cube.coord("longitude").units.name,
    )

    xarr.sic_mean.attrs = dict(
        long_name="mean sea ice area fraction across ensemble runs of icenet "
                  "model",
        standard_name="sea_ice_area_fraction",
        short_name="sic",
        valid_min=0,
        valid_max=1,
        ancillary_variables="sic_stddev",
        grid_mapping="Lambert_Azimuthal_Grid",
        units="1",
    )

    xarr.sic_stddev.attrs = dict(
        long_name="total uncertainty (one standard deviation) of concentration of sea ice",
        standard_name="sea_ice_area_fraction standard_error",
        valid_min=0,
        valid_max=1,
        grid_mapping="Lambert_Azimuthal_Grid",
        units="1",
    )

    # TODO: split into daily files
    output_path = os.path.join(args.output_dir, "{}.nc".format(args.name))
    logging.info("Saving to {}".format(output_path))
    xarr.to_netcdf(output_path)



import logging
import os
import requests
import requests.adapters

from pprint import pformat

import cdsapi as cds
import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.interfaces.downloader import ClimateDownloader
from icenet.data.interfaces.utils import batch_requested_dates

"""
Module to download hourly ERA5 reanalysis latitude-longitude maps,
compute daily averages, regrid them to the same EASE grid as the OSI-SAF sea
ice, data, and save as daily NetCDFs.
"""


class ERA5Downloader(ClimateDownloader):
    """Climate downloader to provide ERA5 reanalysis data from CDS API

    :param identifier: how to identify this dataset
    :param cdi_map: override the default ERA5Downloader.CDI_MAP variable map
    :param use_toolbox: whether to use CDS toolbox for remote aggregation
    :param show_progress: whether to show download progress
    """

    CDI_MAP = {
        'tas': '2m_temperature',
        'ta': 'temperature',  # 500
        'tos': 'sea_surface_temperature',
        'psl': 'surface_pressure',
        'zg': 'geopotential',  # 250 and 500
        'hus': 'specific_humidity',  # 1000
        'rlds': 'surface_thermal_radiation_downwards',
        'rsds': 'surface_solar_radiation_downwards',
        'uas': '10m_u_component_of_wind',
        'vas': '10m_v_component_of_wind',
    }

    def __init__(self,
                 *args,
                 identifier: str = "era5",
                 cdi_map: object = CDI_MAP,
                 use_toolbox: bool = False,
                 show_progress: bool = False,
                 **kwargs):
        super().__init__(*args,
                         identifier=identifier,
                         **kwargs)
        self.client = cds.Client(progress=show_progress)
        self._cdi_map = cdi_map

        self._use_toolbox = use_toolbox
        self.download_method = self._single_api_download

        if use_toolbox:
            self.download_method = self._single_toolbox_download

        if self._max_threads > 10:
            logging.info("Upping connection limit for max_threads > 10")
            adapter = requests.adapters.HTTPAdapter(
                pool_connections=self._max_threads,
                pool_maxsize=self._max_threads
            )
            self.client.session.mount("https://", adapter)

    def _single_toolbox_download(self,
                                 var: object,
                                 level: object,
                                 req_dates: object,
                                 download_path: object):
        """Implements a single download from CDS Toolbox API

        :param var:
        :param level: the pressure level to download
        :param req_dates: the request dates
        :param download_path:
        """

        logging.debug("Processing {} dates".format(len(req_dates)))
        var_prefix = var[0:-(len(str(level)))] if level else var

        params_dict = {
            "realm":    "c3s",
            "project":  "app-c3s-daily-era5-statistics",
            "version":  "master",
            "workflow_name": "application",
            "kwargs": {
                "dataset": "reanalysis-era5-single-levels",
                "product_type": "reanalysis",
                "variable": self._cdi_map[var_prefix],
                "pressure_level": "-",
                "statistic": "daily_mean",
                "year": req_dates[0].year,
                "month": sorted(list(set([r.month for r in req_dates]))),
                "frequency": "1-hourly",
                "time_zone": "UTC+00:00",
                "grid": "0.25/0.25",
                "area": {
                    "lat": [min([self.hemisphere_loc[0],
                                 self.hemisphere_loc[2]]),
                            max([self.hemisphere_loc[0],
                                 self.hemisphere_loc[2]])],
                    "lon": [min([self.hemisphere_loc[1],
                                 self.hemisphere_loc[3]]),
                            max([self.hemisphere_loc[1],
                                 self.hemisphere_loc[3]])],
                },
            },
        }

        if level:
            params_dict["kwargs"]["dataset"] = \
                "reanalysis-era5-pressure-levels"
            params_dict["kwargs"]["pressure_level"] = level

        logging.debug("params_dict: {}".format(pformat(params_dict)))
        result = self.client.service(
            "tool.toolbox.orchestrator.workflow",
            params=params_dict)

        try:
            logging.info("Downloading data for {}...".format(var))
            logging.debug("Result: {}".format(result))
                
            location = result[0]['location']
            res = requests.get(location, stream=True)

            logging.info("Writing data to {}".format(download_path))

            with open(download_path, 'wb') as fh:
                for r in res.iter_content(chunk_size=1024):
                    fh.write(r)

            logging.info("Download completed: {}".format(download_path))

        except Exception as e:
            logging.exception("{} not deleted, look at the "
                              "problem".format(download_path))
            raise RuntimeError(e)

    def _single_api_download(self,
                             var: str,
                             level: object,
                             req_dates: object,
                             download_path: object):
        """Implements a single download from CDS API

        :param var:
        :param level: the pressure level to download
        :param req_dates: the request date
        :param download_path:
        """

        logging.debug("Processing {} dates".format(len(req_dates)))
        var_prefix = var[0:-(len(str(level)))] if level else var

        retrieve_dict = {
            "product_type": "reanalysis",
            "variable": self._cdi_map[var_prefix],
            "year": req_dates[0].year,
            "month": list(set(["{:02d}".format(rd.month)
                               for rd in sorted(req_dates)])),
            "day": ["{:02d}".format(d) for d in range(1, 32)],
            "time": ["{:02d}:00".format(h) for h in range(0, 24)],
            "format": "netcdf",
            "area": self.hemisphere_loc,
        }

        dataset = "reanalysis-era5-single-levels"
        if level:
            dataset = "reanalysis-era5-pressure-levels"
            retrieve_dict["pressure_level"] = level

        try:
            logging.info("Downloading data for {}...".format(var))

            self.client.retrieve(dataset, retrieve_dict, download_path)
            logging.info("Download completed: {}".format(download_path))

        except Exception as e:
            logging.exception("{} not deleted, look at the "
                              "problem".format(download_path))
            raise RuntimeError(e)

    def postprocess(self,
                    var: str,
                    download_path: object):
        """Processing of CDS downloaded files

        If we've not used the toolbox to download the files, we have a lot of
        hourly data to average out, which is taken care of here

        :param var:
        :param download_path:
        """
        # if not self._use_toolbox:
        logging.info("Postprocessing CDS API data at {}".format(download_path))
        da = xr.open_dataarray(download_path)

        if 'expver' in da.coords:
            raise RuntimeError("fix_near_real_time_era5_coords no longer "
                               "exists in the codebase for expver "
                               "in coordinates")

        da = da.resample(time='1D').mean().compute()
        da.to_netcdf(download_path)

    def _get_dates_for_request(self) -> object:
        """Appropriate monthly batching of dates for CDS requests

        :return:

        """
        return batch_requested_dates(self._dates, attribute="month")

    def additional_regrid_processing(self,
                                     datafile: str,
                                     cube_ease: object):
        """

        :param datafile:
        :param cube_ease:
        """
        (datafile_path, datafile_name) = os.path.split(datafile)
        var_name = datafile_path.split(os.sep)[self._var_name_idx]

        # FIXME: are these here or preproc?
        # if var_name == 'zg500' or var_name == 'zg250':
        #   da_daily = da_daily / 9.80665

        # if var_name == 'tos':
        #     # Replace every value outside of SST < 1000 with
        #    zeros (the ERA5 masked values)
        #     da_daily = da_daily.where(da_daily < 1000., 0)
        
        if var_name == 'tos':
            # Overwrite maksed values with zeros
            logging.debug("ERA5 regrid postprocess: {}".format(var_name))
            cube_ease.data[cube_ease.data.mask] = 0.
            cube_ease.data[:, self._masks.get_land_mask()] = 0.
            cube_ease.data = cube_ease.data.data
            cube_ease.data = np.where(np.isnan(cube_ease.data), 0., cube_ease.data)
        elif var_name in ['zg500', 'zg250']:
            # Convert from geopotential to geopotential height
            logging.debug("ERA5 additional regrid: {}".format(var_name))
            cube_ease /= 9.80665


def main():
    args = download_args(choices=["cdsapi", "toolbox"],
                         workers=True, extra_args=(
        (("-n", "--do-not-download"),
         dict(dest="download", action="store_false", default=True)),
        (("-p", "--do-not-postprocess"),
         dict(dest="postprocess", action="store_false", default=True))))

    logging.info("ERA5 Data Downloading")
    era5 = ERA5Downloader(
        var_names=args.vars,
        pressure_levels=args.levels,
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        download=args.download,
        max_threads=args.workers,
        postprocess=args.postprocess,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        use_toolbox=args.choice == "toolbox"
    )
    era5.download()
    era5.regrid()
    era5.rotate_wind_data()



import argparse
import configparser
import logging
import os
import shutil
import tempfile

import xarray as xr

from icenet.process.utils import date_arg, destination_filename
from icenet.utils import setup_logging

from azure.storage.blob import ContainerClient

# https://docs.microsoft.com/en-us/azure/developer/python/sdk/storage/storage-blob-readme?view=storage-py-v12#next-steps


@setup_logging
def upload_parse_args():
    """

    :return:
    """
    a = argparse.ArgumentParser()

    a.add_argument("filename")
    a.add_argument("date", default=None, type=date_arg, nargs="?")

    a.add_argument("-c", "--container", default="input", type=str)
    a.add_argument("-l", "--leave", default=False, action="store_true")
    a.add_argument("-o", "--overwrite", default=False, action="store_true")
    a.add_argument("-v", "--verbose", default=False, action="store_true")

    return a.parse_args()


def upload():
    """

    """
    args = upload_parse_args()
    logging.info("Azure upload facility")

    url = os.getenv("AZURE_STORAGE_CONNECTION_STRING")

    if not url:
        try:
            ini = configparser.RawConfigParser()
            ini.read(os.path.expandvars("$HOME/.icenet.conf"))
            url = ini.get("azure", "connection_string")
        except configparser.Error as e:
            logging.exception("Configuration is not correctly set up")
            raise e

    try:
        if args.date:
            tmpdir = tempfile.mkdtemp(dir=".")
            ds = xr.open_dataset(args.filename)
            ds = ds.sel(time=slice(args.date, args.date))

            if len(ds.time) < 1:
                raise ValueError("No elements in {} for {}".format(
                    args.filename, args.date
                ))

            filename = destination_filename(tmpdir, args.filename, args.date)
            ds.to_netcdf(filename)
            ds.close()
        else:
            filename = args.filename

        with open(filename, "rb") as data:
            logging.info("Uploading {}".format(filename))
            logging.info("Connecting client")

            container_client = \
                ContainerClient.\
                from_connection_string(url, container_name=args.container)
            container_client.upload_blob(
                os.path.basename(filename), data, overwrite=args.overwrite)
    finally:
        if args.date and not args.leave:
            logging.info("Removing {}".format(tmpdir))
            shutil.rmtree(tmpdir)




import argparse
import json
import logging
import os

import numpy as np

from icenet.data.datasets.utils import SplittingMixin
from icenet.data.loader import IceNetDataLoaderFactory
from icenet.data.producers import DataCollection
from icenet.utils import setup_logging

"""


https://stackoverflow.com/questions/55852831/
tf-data-vs-keras-utils-sequence-performance

"""


class IceNetDataSet(SplittingMixin, DataCollection):
    """

    :param configuration_path:
    :param batch_size:
    :param path:
    """

    def __init__(self,
                 configuration_path: str,
                 *args,
                 batch_size: int = 4,
                 path: str = os.path.join(".", "network_datasets"),
                 shuffling: bool = False,
                 **kwargs):
        self._config = dict()
        self._configuration_path = configuration_path
        self._load_configuration(configuration_path)

        super().__init__(*args,
                         identifier=self._config["identifier"],
                         north=bool(self._config["north"]),
                         path=path,
                         south=bool(self._config["south"]),
                         **kwargs)

        self._batch_size = batch_size
        self._counts = self._config["counts"]
        self._dtype = getattr(np, self._config["dtype"])
        self._loader_config = self._config["loader_config"]
        self._n_forecast_days = self._config["n_forecast_days"]
        self._num_channels = self._config["num_channels"]
        self._shape = tuple(self._config["shape"])
        self._shuffling = shuffling

        if self._config["loader_path"] and \
                os.path.exists(self._config["loader_path"]):
            hemi = self.hemisphere_str[0]
            self.add_records(self.base_path, hemi)
        else:
            logging.warning("Running in configuration only mode, tfrecords "
                            "were not generated for this dataset")

    def _load_configuration(self, path: str):
        """

        :param path:
        """
        if os.path.exists(path):
            logging.info("Loading configuration {}".format(path))

            with open(path, "r") as fh:
                obj = json.load(fh)

                self._config.update(obj)
        else:
            raise OSError("{} not found".format(path))

    def get_data_loader(self):
        """

        :return:
        """
        loader = IceNetDataLoaderFactory().create_data_loader(
            "dask",
            self.loader_config,
            self.identifier,
            self._config["var_lag"],
            dataset_config_path=os.path.dirname(
                self._configuration_path),
            loss_weight_days=self._config[
                "loss_weight_days"],
            north=self.north,
            output_batch_size=self._config[
                "output_batch_size"],
            south=self.south,
            var_lag_override=self._config[
                "var_lag_override"])
        return loader

    @property
    def loader_config(self):
        return self._loader_config

    @property
    def channels(self):
        return self._config["channels"]

    @property
    def counts(self):
        return self._config["counts"]


class MergedIceNetDataSet(SplittingMixin, DataCollection):
    """

    :param identifier:
    :param configuration_paths: List of configurations to load
    :param batch_size:
    :param path:
    """

    def __init__(self,
                 configuration_paths: object,
                 *args,
                 batch_size: int = 4,
                 path: str = os.path.join(".", "network_datasets"),
                 shuffling: bool = False,
                 **kwargs):
        self._config = dict()
        self._configuration_paths = [configuration_paths] \
            if type(configuration_paths) != list else configuration_paths
        self._load_configurations(configuration_paths)

        identifier = ".".join([loader.identifier
                               for loader in self._config["loaders"]])

        super().__init__(*args,
                         identifier=identifier,
                         north=bool(self._config["north"]),
                         path=path,
                         south=bool(self._config["south"]),
                         **kwargs)

        self._base_path = path
        self._batch_size = batch_size
        self._dtype = getattr(np, self._config["dtype"])
        self._num_channels = self._config["num_channels"]
        self._n_forecast_days = self._config["n_forecast_days"]
        self._shape = self._config["shape"]
        self._shuffling = shuffling

        self._init_records()

    def _init_records(self):
        """

        """
        for idx, loader_path in enumerate(self._config["loader_paths"]):
            hemi = self._config["loaders"][idx].hemisphere_str[0]
            base_path = os.path.join(self._base_path,
                                     self._config["loaders"][idx].identifier)
            self.add_records(base_path, hemi)

    def _load_configurations(self, paths: object):
        """

        :param paths:
        """
        self._config = dict(
            loader_paths=[],
            loaders=[],
            north=False,
            south=False
        )
        
        for path in paths:
            if os.path.exists(path):
                logging.info("Loading configuration {}".format(path))

                with open(path, "r") as fh:
                    obj = json.load(fh)
                    self._merge_configurations(path, obj)
            else:
                raise OSError("{} not found".format(path))

    def _merge_configurations(self, path: str, other: object):
        """

        :param path:
        :param other:
        """
        loader = IceNetDataLoaderFactory().create_data_loader(
            "dask",
            other["loader_config"],
            other["identifier"],
            other["var_lag"],
            dataset_config_path=os.path.dirname(path),
            loss_weight_days=other["loss_weight_days"],
            north=other["north"],
            output_batch_size=other["output_batch_size"],
            south=other["south"],
            var_lag_override=other["var_lag_override"])

        self._config["loaders"].append(loader)
        self._config["loader_paths"].append(other["loader_path"])

        if "counts" not in self._config:
            self._config["counts"] = other["counts"].copy()
        else:
            for dataset, count in other["counts"].items():
                logging.info("Merging {} samples from {}".format(count, dataset))
                self._config["counts"][dataset] += count

        general_attrs = ["channels", "dtype", "n_forecast_days",
                         "num_channels", "output_batch_size", "shape"]

        for attr in general_attrs:
            if attr not in self._config:
                self._config[attr] = other[attr]
            else:
                assert self._config[attr] == other[attr], \
                    "{} is not the same across configurations".format(attr)

        self._config["north"] = True if loader.north else self._config["north"]
        self._config["south"] = True if loader.south else self._config["south"]

    def get_data_loader(self):
        """

        :return:
        """
        assert len(self._configuration_paths) == 1, "Configuration mode is " \
                                                    "only for single loader" \
                                                    "datasets: {}".format(
            self._configuration_paths
        )
        return self._config["loader"][0]

    def check_dataset(self,
                      split: str = "train"):
        """

        :param split:
        """
        raise NotImplementedError("Checking not implemented for merged sets, "
                                  "consider doing them individually")

    @property
    def channels(self):
        return self._config['channels']

    @property
    def counts(self):
        return self._config["counts"]


@setup_logging
def get_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("dataset")
    ap.add_argument("-s", "--split",
                    choices=["train", "val", "test"], default="train")
    ap.add_argument("-v", "--verbose", action="store_true", default=False)
    args = ap.parse_args()
    return args


def check_dataset():
    args = get_args()
    ds = IceNetDataSet(args.dataset)
    ds.check_dataset(args.split)



import logging
import requests

import cartopy.crs as ccrs
import iris
import numpy as np


def assign_lat_lon_coord_system(cube: object):
    """Assign coordinate system to iris cube to allow regridding.

    :param cube:
    """

    # This originated from era5
    cube.coord('latitude').coord_system = iris.coord_systems.GeogCS(6367470.0)
    cube.coord('longitude').coord_system = iris.coord_systems.GeogCS(6367470.0)

    # NOTE: CMIP6 original assignment, but cs in this case is bring assigned to
    # a module call that doesn't exist, let alone contain this member
    # cs = grid_cube.coord_system().ellipsoid
    # for coord in ['longitude', 'latitude']:
    #     cmip6_cube.coord(coord).coord_system = cs

    return cube


def rotate_grid_vectors(u_cube: object,
                        v_cube: object,
                        angles: object):
    """
    Author: Tony Phillips (BAS)

    Wrapper for :func:`~iris.analysis.cartography.rotate_grid_vectors`
    that can rotate multiple masked spatial fields in one go by iterating
    over the horizontal spatial axes in slices

    :param u_cube:
    :param v_cube:
    :param angles:
    :return:

    """
    # lists to hold slices of rotated vectors
    u_r_all = iris.cube.CubeList()
    v_r_all = iris.cube.CubeList()

    # get the X and Y dimension coordinates for each source cube
    u_xy_coords = [u_cube.coord(axis='x', dim_coords=True),
                   u_cube.coord(axis='y', dim_coords=True)]
    v_xy_coords = [v_cube.coord(axis='x', dim_coords=True),
                   v_cube.coord(axis='y', dim_coords=True)]

    # iterate over X, Y slices of the source cubes, rotating each in turn
    for u, v in zip(u_cube.slices(u_xy_coords, ordered=False),
                    v_cube.slices(v_xy_coords, ordered=False)):
        u_r, v_r = iris.analysis.cartography.rotate_grid_vectors(u, v, angles)
        u_r_all.append(u_r)
        v_r_all.append(v_r)

    # return the slices, merged back together into a pair of cubes
    return u_r_all.merge_cube(), v_r_all.merge_cube()


def gridcell_angles_from_dim_coords(cube: object):
    """
    Author: Tony Phillips (BAS)

    Wrapper for :func:`~iris.analysis.cartography.gridcell_angles`
    that derives the 2D X and Y lon/lat coordinates from 1D X and Y
    coordinates identifiable as 'x' and 'y' axes

    The provided cube must have a coordinate system so that its
    X and Y coordinate bounds (which are derived if necessary)
    can be converted to lons and lats

    :param cube:
    :return:
    """

    # get the X and Y dimension coordinates for the cube
    x_coord = cube.coord(axis='x', dim_coords=True)
    y_coord = cube.coord(axis='y', dim_coords=True)

    # add bounds if necessary
    if not x_coord.has_bounds():
        x_coord = x_coord.copy()
        x_coord.guess_bounds()
    if not y_coord.has_bounds():
        y_coord = y_coord.copy()
        y_coord.guess_bounds()

    # get the grid cell bounds
    x_bounds = x_coord.bounds
    y_bounds = y_coord.bounds
    nx = x_bounds.shape[0]
    ny = y_bounds.shape[0]

    # make arrays to hold the ordered X and Y bound coordinates
    x = np.zeros((ny, nx, 4))
    y = np.zeros((ny, nx, 4))

    # iterate over the bounds (in order BL, BR, TL, TR), mesh them and
    # put them into the X and Y bound coordinates (in order BL, BR, TR, TL)
    c = [0, 1, 3, 2]
    cind = 0
    for yi in [0, 1]:
        for xi in [0, 1]:
            xy = np.meshgrid(x_bounds[:, xi], y_bounds[:, yi])
            x[:,:,c[cind]] = xy[0]
            y[:,:,c[cind]] = xy[1]
            cind += 1

    # convert the X and Y coordinates to longitudes and latitudes
    source_crs = cube.coord_system().as_cartopy_crs()
    target_crs = ccrs.PlateCarree()
    pts = target_crs.transform_points(source_crs, x.flatten(), y.flatten())
    lons = pts[:, 0].reshape(x.shape)
    lats = pts[:, 1].reshape(x.shape)

    # get the angles
    angles = iris.analysis.cartography.gridcell_angles(lons, lats)

    # add the X and Y dimension coordinates from the cube to the angles cube
    angles.add_dim_coord(y_coord, 0)
    angles.add_dim_coord(x_coord, 1)

    # if the cube's X dimension preceeds its Y dimension
    # transpose the angles to match
    if cube.coord_dims(x_coord)[0] < cube.coord_dims(y_coord)[0]:
        angles.transpose()

    return angles


def invert_gridcell_angles(angles: object):
    """
    Author: Tony Phillips (BAS)

    Negate a cube of gridcell angles in place, transforming
    gridcell_angle_from_true_east <--> true_east_from_gridcell_angle
    :param angles:
    """
    angles.data *= -1

    names = ['true_east_from_gridcell_angle', 'gridcell_angle_from_true_east']
    name = angles.name()
    if name in names:
        angles.rename(names[1 - names.index(name)])


def esgf_search(server: str = "https://esgf-node.llnl.gov/esg-search/search",
                files_type: str = "OPENDAP",
                local_node: bool = False,
                latest: bool = True,
                project: str = "CMIP6",
                format: str = "application%2Fsolr%2Bjson",
                use_csrf: bool = False,
                **search):
    """

    Below taken from
    https://hub.binder.pangeo.io/user/pangeo-data-pan--cmip6-examples-ro965nih/lab
    and adapted slightly

    :param server:
    :param files_type:
    :param local_node:
    :param latest:
    :param project:
    :param format:
    :param use_csrf:
    :param search:
    :return:
    """
    client = requests.session()
    payload = search
    payload["project"] = project
    payload["type"]="File"
    if latest:
        payload["latest"] = "true"
    if local_node:
        payload["distrib"] = "false"
    if use_csrf:
        client.get(server)
        if 'csrftoken' in client.cookies:
            # Django 1.6 and up
            csrftoken = client.cookies['csrftoken']
        else:
            # older versions
            csrftoken = client.cookies['csrf']
        payload["csrfmiddlewaretoken"] = csrftoken

    payload["format"] = format

    offset = 0
    numFound = 10000
    all_files = []
    files_type = files_type.upper()
    while offset < numFound:
        payload["offset"] = offset
        url_keys = []
        for k in payload:
            url_keys += ["{}={}".format(k, payload[k])]

        url = "{}/?{}".format(server, "&".join(url_keys))
        logging.debug("ESGF search URL: {}".format(url))

        r = client.get(url)
        r.raise_for_status()
        resp = r.json()["response"]
        numFound = int(resp["numFound"])
        resp = resp["docs"]
        offset += len(resp)
        for d in resp:
            for k in d:
                logging.debug("{}: {}".format(k,d[k]))
            url = d["url"]
            for f in d["url"]:
                sp = f.split("|")
                if sp[-1] == files_type:
                    all_files.append(sp[0].split(".html")[0])
    return sorted(all_files)



import argparse
import logging
import os

from datetime import timedelta

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.animation import FuncAnimation
matplotlib.use('Agg')

import numpy as np
import pandas as pd

from icenet.data.cli import date_arg
from icenet.data.sic.mask import Masks
from icenet.plotting.utils import \
    filter_ds_by_obs, get_forecast_ds, get_obs_da, get_seas_forecast_da

matplotlib.rcParams.update({
    'figure.facecolor': 'w',
    'figure.dpi': 300
})


def plot_binary_accuracy(masks: object,
                         fc_da: object,
                         cmp_da: object,
                         obs_da: object,
                         output_path: object =
                         os.path.join("plot", "binacc.png")) -> object:
    """

    :param masks:
    :param fc_da:
    :param cmp_da:
    :param obs_da:
    :param output_path:
    :return:
    """
    agcm = masks.get_active_cell_da(obs_da)
    binary_obs_da = obs_da > 0.15

    fig, ax = plt.subplots(figsize=(12, 6))
    ax.set_title("Binary accuracy comparison")

    binary_fc_da = fc_da > 0.15
    binary_fc_da = (binary_fc_da == binary_obs_da).\
        astype(np.float16).weighted(agcm)
    binacc_fc = (binary_fc_da.mean(dim=['yc', 'xc']) * 100)
    ax.plot(binacc_fc.time, binacc_fc.values, label="IceNet")

    if cmp_da is not None:
        binary_cmp_da = cmp_da > 0.15
        binary_cmp_da = (binary_cmp_da == binary_obs_da).\
            astype(np.float16).weighted(agcm)
        binacc_cmp = (binary_cmp_da.mean(dim=['yc', 'xc']) * 100)
        ax.plot(binacc_cmp.time, binacc_cmp.values, label="SEAS")
    else:
        binacc_cmp = None

    ax.xaxis.set_major_formatter(
        mdates.ConciseDateFormatter(ax.xaxis.get_major_locator()))
    ax.xaxis.set_major_locator(mdates.MonthLocator())
    ax.xaxis.set_minor_locator(mdates.DayLocator())
    ax.legend(loc='lower right')

    plt.savefig(output_path)

    return binacc_fc, binacc_cmp


def sic_error_video(fc_da: object,
                    obs_da: object,
                    land_mask: object,
                    output_path: object =
                    os.path.join("plot", "sic_error.mp4")) -> object:
    """

    :param fc_da:
    :param obs_da:
    :param land_mask:
    :param output_path:
    """

    diff = fc_da - obs_da

    fig, maps = plt.subplots(nrows=1, ncols=3, figsize=(18, 5))
    fig.set_dpi(150)

    leadtime = 0
    fc_plot = fc_da.isel(time=leadtime).to_numpy()
    obs_plot = obs_da.isel(time=leadtime).to_numpy()
    diff_plot = diff.isel(time=leadtime).to_numpy()

    contour_kwargs = dict(
        vmin=0,
        vmax=1,
        cmap='YlOrRd'
    )

    im1 = maps[0].imshow(fc_plot, **contour_kwargs)
    im2 = maps[1].imshow(obs_plot, **contour_kwargs)
    im3 = maps[2].imshow(diff_plot, 
                         vmin=-1, vmax=1, cmap="RdBu_r")

    tic = maps[0].set_title("IceNet {}".format(
        pd.to_datetime(fc_da.isel(time=leadtime).time.values).strftime("%d/%m/%Y")))
    tio = maps[1].set_title("OSISAF Obs {}".format(
        pd.to_datetime(obs_da.isel(time=leadtime).time.values).strftime("%d/%m/%Y")))
    maps[2].set_title("Diff")

    p0 = maps[0].get_position().get_points().flatten()
    p1 = maps[1].get_position().get_points().flatten()
    p2 = maps[2].get_position().get_points().flatten()

    ax_cbar = fig.add_axes([p0[0], 0, p1[2]-p0[0], 0.05])
    plt.colorbar(im1, cax=ax_cbar, orientation='horizontal')

    ax_cbar1 = fig.add_axes([p2[0], 0, p2[2]-p2[0], 0.05])
    plt.colorbar(im3, cax=ax_cbar1, orientation='horizontal')

    for m_ax in maps[0:3]:
        m_ax.contourf(land_mask,
                      levels=[.5, 1],
                      colors=[matplotlib.cm.gray(180)],
                      zorder=3)

    fig.subplots_adjust(hspace=0.2, wspace=0.2)

    def update(date):
        logging.debug("Plotting {}".format(date))

        fc_plot = fc_da.isel(time=date).to_numpy()
        obs_plot = obs_da.isel(time=date).to_numpy()
        diff_plot = diff.isel(time=date).to_numpy()
        
        tic.set_text("IceNet {}".format(
            pd.to_datetime(fc_da.isel(time=date).time.values).strftime("%d/%m/%Y")))
        tio.set_text("OSISAF Obs {}".format(
            pd.to_datetime(obs_da.isel(time=date).time.values).strftime("%d/%m/%Y")))

        im1.set_data(fc_plot)
        im2.set_data(obs_plot)
        im3.set_data(diff_plot)

        return tic, tio, im1, im2, im3

    animation = FuncAnimation(fig,
                              update,
                              range(0, len(fc_da.time)),
                              interval=100)

    plt.close()

    logging.info("Saving plot to {}".format(output_path))
    animation.save(output_path,
                   fps=10,
                   extra_args=['-vcodec', 'libx264'])
    return animation


def forecast_plot_args(ecmwf: bool = True) -> object:
    """

    :return:
    """

    ap = argparse.ArgumentParser()
    ap.add_argument("hemisphere", choices=("north", "south"))
    ap.add_argument("forecast_file", type=str)
    ap.add_argument("forecast_date", type=date_arg)

    ap.add_argument("-o", "--output-path", type=str, default=None)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    if ecmwf:
        ap.add_argument("-e", "--ecmwf", action="store_true", default=False)

    args = ap.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)
    logging.getLogger("matplotlib").setLevel(logging.WARNING)

    return args


def binary_accuracy():
    """

    """
    args = forecast_plot_args()

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    seas = get_seas_forecast_da(args.hemisphere, args.forecast_date) \
        if args.ecmwf else None

    seas = seas.assign_coords(dict(xc=seas.xc / 1e3, yc=seas.yc / 1e3))

    plot_binary_accuracy(masks,
                         fc,
                         seas.isel(time=slice(1, None)),
                         obs,
                         args.output_path)


def sic_error():
    """

    """
    args = forecast_plot_args(ecmwf=False)

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    sic_error_video(fc,
                    obs,
                    masks.get_land_mask(),
                    args.output_path)



import argparse
import glob
import logging
import os
import sys

import numpy as np
import pandas as pd
import xarray as xr

from icenet.utils import Hemisphere
from icenet.data.producers import DataProducer

from scipy import interpolate
from scipy.spatial.qhull import QhullError

"""

"""


def sic_interpolate(da: object,
                    masks: object) -> object:
    """

    :param da:
    :param masks:
    :return:
    """
    for date in da.time.values:
        polarhole_mask = masks.get_polarhole_mask(
            pd.to_datetime(date).date())

        da_day = da.sel(time=date)
        xx, yy = np.meshgrid(np.arange(432), np.arange(432))

        # Grid cells outside of polar hole or NaN regions
        valid = ~np.isnan(da_day.data)

        # Interpolate polar hole
        if type(polarhole_mask) is np.ndarray:
            valid = valid & ~polarhole_mask

        # Interpolate if there is more than one missing grid cell
        if np.sum(~valid) >= 1:
            # Find grid cell locations surrounding NaN regions for bilinear
            # interpolation
            nan_mask = np.ma.masked_array(np.full((432, 432), 0.))
            nan_mask[~valid] = np.ma.masked

            nan_neighbour_arrs = {}
            for order in 'C', 'F':
                # starts and ends indexes of masked element chunks
                slice_ends = np.ma.clump_masked(nan_mask.ravel(order=order))

                nan_neighbour_idxs = []
                nan_neighbour_idxs.extend([s.start for s in slice_ends])
                nan_neighbour_idxs.extend([s.stop - 1 for s in slice_ends])

                nan_neighbour_arr_i = np.array(np.full((432, 432), False),
                                               order=order)
                nan_neighbour_arr_i.ravel(order=order)[nan_neighbour_idxs] \
                    = True
                nan_neighbour_arrs[order] = nan_neighbour_arr_i

            nan_neighbour_arr = nan_neighbour_arrs['C'] + \
                nan_neighbour_arrs['F']
            # Remove artefacts along edge of the grid
            nan_neighbour_arr[:, 0] = \
                nan_neighbour_arr[0, :] = \
                nan_neighbour_arr[:, -1] = \
                nan_neighbour_arr[-1, :] = False

            if np.sum(nan_neighbour_arr) == 1:
                res = np.where(np.array(nan_neighbour_arr) == True)
                logging.warning("Not enough nans for interpolation, extending {}".format(res))
                x_idx, y_idx = res[0][0], res[1][0]
                nan_neighbour_arr[x_idx-1:x_idx+2, y_idx] = True
                nan_neighbour_arr[x_idx, y_idx-1:y_idx+2] = True
                logging.debug(np.where(np.array(nan_neighbour_arr) == True))

            # Perform bilinear interpolation
            x_valid = xx[nan_neighbour_arr]
            y_valid = yy[nan_neighbour_arr]
            values = da_day.data[nan_neighbour_arr]

            x_interp = xx[~valid]
            y_interp = yy[~valid]

            try:
                if len(x_valid) or len(y_valid):
                    interp_vals = interpolate.griddata((x_valid, y_valid),
                                                       values,
                                                       (x_interp, y_interp),
                                                       method='linear')
                    da.sel(time=date).data[~valid] = interp_vals
                else:
                    logging.warning("No valid values to interpolate with on "
                                    "{}".format(date))
            except QhullError:
                logging.exception("Geometrical degeneracy from QHull, interpolation failed")

    return da


def condense_main():
    ap = argparse.ArgumentParser()
    ap.add_argument("identifier")
    ap.add_argument("hemisphere", choices=("north", "south"))
    ap.add_argument("variable")

    ap.add_argument("-n", "--numpy", action="store_true", default=False)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)
    args = ap.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)
    condense_data(args.identifier, args.hemisphere, args.variable)


def condense_data(identifier: str,
                  hemisphere: str,
                  variable: str):
    """Takes existing daily files and creates yearly files

    Previous early versions of the pipeline were storing files day by day, which
    is pretty wasteful. This allows us to create the yearly files and avoid
    all that nasty re-downloading business

    :param identifier:
    :param hemisphere:
    :param variable:
    """
    logging.info("Condensing data into singular file")

    dp = DataProducer(identifier=identifier,
                      north=getattr(Hemisphere,
                                    hemisphere.upper()) == Hemisphere.NORTH,
                      south=getattr(Hemisphere,
                                    hemisphere.upper()) == Hemisphere.SOUTH)

    data_path = dp.get_data_var_folder(variable, missing_error=True)

    logging.debug("Collecting files from {}".format(data_path))
    dfs = glob.glob(os.path.join(data_path, "**", "*.nc"))

    def year_batch(filenames):
        df_years = set([os.path.split(os.path.dirname(f_year))[-1]
                        for f_year in filenames])

        for year_el in df_years:
            year_dfs = [el for el in filenames
                        if os.path.split(os.path.dirname(el))[-1] == year_el
                        and not os.path.split(el)[1].startswith("latlon")]
            logging.debug("{} has {} files".format(year_el, len(year_dfs)))
            yield year_el, year_dfs

    if len(dfs):
        logging.debug("Got {} files, collecting to {}...".format(len(dfs),
                                                                 data_path))

        for year, year_files in year_batch(dfs):
            year_path = os.path.join(data_path, "{}.nc".format(year))

            if not os.path.exists(year_path):
                logging.info("Loading {}".format(year))
                ds = xr.open_mfdataset(year_files, parallel=True)
                years, datasets = zip(*ds.groupby("time.year"))
                if len(years) > 1:
                    raise RuntimeError("Too many years in one file {}".
                                       format(years))
                logging.info("Saving to {}".format(year_path))
                xr.save_mfdataset(datasets, [year_path])
    else:
        logging.info("No valid files found.")



import datetime as dt
import json
import logging
import os
from abc import abstractmethod

from pprint import pformat

import numpy as np

from icenet.data.process import IceNetPreProcessor
from icenet.data.producers import Generator

"""

"""


class IceNetBaseDataLoader(Generator):
    """

    :param configuration_path,
    :param identifier,
    :param var_lag,
    :param dataset_config_path:
    :param generate_workers:
    :param loss_weight_days:
    :param n_forecast_days:
    :param output_batch_size:
    :param path:
    :param var_lag_override:
    """

    def __init__(self,
                 configuration_path: str,
                 identifier: str,
                 var_lag: int,
                 *args,
                 dataset_config_path: str = ".",
                 dates_override: object = None,
                 dry: bool = False,
                 generate_workers: int = 8,
                 loss_weight_days: bool = True,
                 n_forecast_days: int = 93,
                 output_batch_size: int = 32,
                 path: str = os.path.join(".", "network_datasets"),
                 pickup: bool = False,
                 var_lag_override: object = None,
                 **kwargs):
        super().__init__(*args,
                         identifier=identifier,
                         path=path,
                         **kwargs)

        self._channels = dict()
        self._channel_files = dict()

        self._configuration_path = configuration_path
        self._dataset_config_path = dataset_config_path
        self._dates_override = dates_override
        self._config = dict()
        self._dry = dry
        self._loss_weight_days = loss_weight_days
        self._meta_channels = []
        self._missing_dates = []
        self._n_forecast_days = n_forecast_days
        self._output_batch_size = output_batch_size
        self._pickup = pickup
        self._trend_steps = dict()
        self._workers = generate_workers

        self._var_lag = var_lag
        self._var_lag_override = dict() \
            if not var_lag_override else var_lag_override

        self._load_configuration(configuration_path)
        self._construct_channels()

        self._dtype = getattr(np, self._config["dtype"])
        self._shape = tuple(self._config["shape"])

        self._missing_dates = [
            dt.datetime.strptime(s, IceNetPreProcessor.DATE_FORMAT)
            for s in self._config["missing_dates"]]

    def write_dataset_config_only(self):
        """

        """
        splits = ("train", "val", "test")
        counts = {el: 0 for el in splits}

        logging.info("Writing dataset configuration without data generation")

        # FIXME: cloned mechanism from generate() - do we need to treat these as
        #  sets that might have missing data for fringe cases?
        for dataset in splits:
            forecast_dates = sorted(list(set(
                [dt.datetime.strptime(s,
                 IceNetPreProcessor.DATE_FORMAT).date()
                 for identity in
                 self._config["sources"].keys()
                 for s in
                 self._config["sources"][identity]
                 ["dates"][dataset]])))

            logging.info("{} {} dates in total, NOT generating cache "
                         "data.".format(len(forecast_dates), dataset))
            counts[dataset] += len(forecast_dates)

        self._write_dataset_config(counts, network_dataset=False)

    @abstractmethod
    def generate_sample(self,
                        date: object,
                        prediction: bool = False):
        """

        :param date:
        :param prediction:
        :return:
        """
        pass

    def get_sample_files(self) -> object:
        """

        :param date:
        :return:
        """
        # FIXME: is this not just the same as _channel_files now?
        # FIXME: still experimental code, move to multiple implementations
        # FIXME: CLEAN THIS ALL UP ONCE VERIFIED FOR local/shared STORAGE!
        var_files = dict()

        for var_name, num_channels in self._channels.items():
            var_file = self._get_var_file(var_name)

            if not var_file:
                raise RuntimeError("No file returned for {}".format(var_name))

            if var_name not in var_files:
                var_files[var_name] = var_file
            elif var_file != var_files[var_name]:
                raise RuntimeError("Differing files? {} {} vs {}".
                                   format(var_name,
                                          var_file,
                                          var_files[var_name]))

        return var_files

    def _add_channel_files(self,
                           var_name: str,
                           filelist: object):
        """

        :param var_name:
        :param filelist:
        """
        if var_name in self._channel_files:
            logging.warning("{} already has files, but more found, "
                            "this could be an unintentional merge of "
                            "sources".format(var_name))
        else:
            self._channel_files[var_name] = []

        logging.debug("Adding {} to {} channel".format(len(filelist), var_name))
        self._channel_files[var_name] += filelist

    def _construct_channels(self):
        """

        """
        # As of Python 3.7 dict guarantees the order of keys based on
        # original insertion order, which is great for this method
        lag_vars = [(identity, var, data_format)
                    for data_format in ("abs", "anom")
                    for identity in
                    sorted(self._config["sources"].keys())
                    for var in
                    sorted(self._config["sources"][identity][data_format])]

        for identity, var_name, data_format in lag_vars:
            var_prefix = "{}_{}".format(var_name, data_format)
            var_lag = (self._var_lag
                       if var_name not in self._var_lag_override
                       else self._var_lag_override[var_name])

            self._channels[var_prefix] = int(var_lag)
            self._add_channel_files(
                var_prefix,
                [el for el in
                 self._config["sources"][identity]["var_files"][var_name]
                 if var_prefix in os.path.split(el)[1]])

        trend_names = [(identity, var,
                        self._config["sources"][identity]["linear_trend_steps"])
                       for identity in
                       sorted(self._config["sources"].keys())
                       for var in
                       sorted(
                           self._config["sources"][identity]["linear_trends"])]

        for identity, var_name, trend_steps in trend_names:
            var_prefix = "{}_linear_trend".format(var_name)

            self._channels[var_prefix] = len(trend_steps)
            self._trend_steps[var_prefix] = trend_steps
            filelist = [el for el in
                        self._config["sources"][identity]["var_files"][var_name]
                        if "linear_trend" in os.path.split(el)[1]]

            self._add_channel_files(var_prefix, filelist)

        # Metadata input variables that don't span time
        meta_names = [(identity, var)
                      for identity in
                      sorted(self._config["sources"].keys())
                      for var in
                      sorted(self._config["sources"][identity]["meta"])]

        for identity, var_name in meta_names:
            self._meta_channels.append(var_name)
            self._channels[var_name] = 1
            self._add_channel_files(
                var_name,
                self._config["sources"][identity]["var_files"][var_name])

        logging.debug("Channel quantities deduced:\n{}\n\nTotal channels: {}".
                      format(pformat(self._channels), self.num_channels))

    def _get_var_file(self, var_name: str):
        """

        :param var_name:
        :return:
        """

        filename = "{}.nc".format(var_name)
        files = self._channel_files[var_name]

        if len(self._channel_files[var_name]) > 1:
            logging.warning("Multiple files found for {}, only returning {}".
                            format(filename, files[0]))
        elif not len(files):
            logging.warning("No files in channel list for {}".format(filename))
            return None
        return files[0]

    def _load_configuration(self, path: str):
        """

        :param path:
        """
        if os.path.exists(path):
            logging.info("Loading configuration {}".format(path))

            with open(path, "r") as fh:
                obj = json.load(fh)

                self._config.update(obj)
        else:
            raise OSError("{} not found".format(path))

    def _write_dataset_config(self,
                              counts: object,
                              network_dataset: bool = True):
        """

        :param counts:
        :param network_dataset:
        :return:
        """
        # TODO: move to utils for this and process
        def _serialize(x):
            if x is dt.date:
                return x.strftime(IceNetPreProcessor.DATE_FORMAT)
            return str(x)

        configuration = {
            "identifier":       self.identifier,
            "implementation":   self.__class__.__name__,
            # This is only for convenience ;)
            "channels":         [
                "{}_{}".format(channel, i)
                for channel, s in
                self._channels.items()
                for i in range(1, s + 1)],
            "counts":           counts,
            "dtype":            self._dtype.__name__,
            "loader_config":    self._configuration_path,
            "missing_dates":    [date.strftime(
                IceNetPreProcessor.DATE_FORMAT) for date in
                self._missing_dates],
            "n_forecast_days":  self._n_forecast_days,
            "north":            self.north,
            "num_channels":     self.num_channels,
            # FIXME: this naming is inconsistent, sort it out!!! ;)
            "shape":            list(self._shape),
            "south":            self.south,

            # For recreating this dataloader
            # "dataset_config_path = ".",
            # FIXME: badly named, should really be dataset_path
            "loader_path":      self._path if network_dataset else False,
            "loss_weight_days": self._loss_weight_days,
            "output_batch_size": self._output_batch_size,
            "var_lag":          self._var_lag,
            "var_lag_override": self._var_lag_override,
        }

        output_path = os.path.join(self._dataset_config_path,
                                   "dataset_config.{}.json".format(
                                       self.identifier))

        logging.info("Writing configuration to {}".format(output_path))

        with open(output_path, "w") as fh:
            json.dump(configuration, fh, indent=4, default=_serialize)

    @property
    def config(self):
        return self._config

    @property
    def dates_override(self):
        return self._dates_override

    @property
    def num_channels(self):
        return sum(self._channels.values())

    @property
    def pickup(self):
        return self._pickup

    @property
    def workers(self):
        return self._workers





import tensorflow as tf


"""

"""


class IceNetDataWarning(RuntimeWarning):
    pass


def write_tfrecord(writer: object,
                   x: object,
                   y: object,
                   sample_weights: object):
    """

    :param writer:
    :param x:
    :param y:
    :param sample_weights:
    :param data_check:
    """

    # FIXME: this will trigger eager computation of the dataset, should be
    #  optional but for the moment is commented out. It's potentially better
    #  situated elsewhere too
    #
    #    y_nans = da.isnan(y).sum()
    #    x_nans = da.isnan(x).sum()
    #    sw_nans = da.isnan(sample_weights).sum()

    #    if y_nans + x_nans + sw_nans > 0:
    #        logging.warning("NaNs detected {}: input = {}, "
    #                        "output = {}, weights = {}".
    #                        format(forecast_date, x_nans, y_nans, sw_nans))

    #        if data_check and sample_weights[da.isnan(y)].sum() > 0:
    #            raise IceNetDataWarning("NaNs in output with non-zero weights")

    #        if data_check and x_nans > 0:

    record_data = tf.train.Example(features=tf.train.Features(feature={
        "x": tf.train.Feature(
            float_list=tf.train.FloatList(value=x.reshape(-1))),
        "y": tf.train.Feature(
            float_list=tf.train.FloatList(value=y.reshape(-1))),
        "sample_weights": tf.train.Feature(
            float_list=tf.train.FloatList(value=sample_weights.reshape(-1))),
    })).SerializeToString()

    writer.write(record_data)



import argparse
import collections
import datetime as dt
import logging
import re

from pprint import pformat

import pandas as pd

from icenet.utils import setup_logging

"""

"""


def date_arg(string: str) -> object:
    """

    :param string: 
    :return: 
    """
    date_match = re.search(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)
    return dt.date(*[int(s) for s in date_match.groups()])


def dates_arg(string: str) -> object:
    """

    :param string: 
    :return: 
    """
    if string == "none":
        return []

    date_match = re.findall(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)

    if len(date_match) < 1:
        raise argparse.ArgumentError("No dates found for supplied argument {}".
                                     format(string))
    return [dt.date(*[int(s) for s in date_tuple]) for date_tuple in date_match]


def csv_arg(string: str) -> list:
    """

    :param string:
    :return:
    """
    csv_items = []
    for el in string.split(","):
        if len(el) == 0:
            csv_items.append(None)
        else:
            csv_items.append(el)
    return csv_items


def csv_of_csv_arg(string: str) -> list:
    """

    :param string:
    :return:
    """
    csv_items = []
    for el in string.split(","):
        if len(el) == 0:
            csv_items.append(None)
        else:
            csv_items.append(el.split("|"))
    return csv_items


def int_or_list_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    try:
        val = int(string)
    except ValueError:
        val = string.split(",")
    return val


@setup_logging
def download_args(choices: object = None,
                  dates: bool = True,
                  dates_optional: bool = False,
                  var_specs: bool = True,
                  workers: bool = False,
                  extra_args: object = ()) -> object:
    """

    :param choices:
    :param dates:
    :param dates_optional:
    :param var_specs:
    :param workers:
    :param extra_args:
    :return:
    """

    ap = argparse.ArgumentParser()
    ap.add_argument("hemisphere", choices=("north", "south"))

    if choices and type(choices) == list:
        ap.add_argument("-c", "--choice", choices=choices, default=choices[0])

    if dates:
        pos_args = [["start_date"], ["end_date"]] if not dates_optional else \
            [["-sd", "--start-date"], ["-ed", "--end-date"]]
        ap.add_argument(*pos_args[0], type=date_arg, default=None)
        ap.add_argument(*pos_args[1], type=date_arg, default=None)

    if workers:
        ap.add_argument("-w", "--workers", default=8, type=int)

    ap.add_argument("-d", "--dont-delete", dest="delete",
                    action="store_false", default=True)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    if var_specs:
        ap.add_argument("--vars",
                        help="Comma separated list of vars",
                        type=csv_arg,
                        default=[])
        ap.add_argument("--levels",
                        help="Comma separated list of pressures/depths as needed, "
                             "use zero length string if None (e.g. ',,500,,,') and "
                             "pipes for multiple per var (e.g. ',,250|500,,'",
                        type=csv_of_csv_arg,
                        default=[])

    for arg in extra_args:
        ap.add_argument(*arg[0], **arg[1])
    args = ap.parse_args()
    return args


@setup_logging
def process_args(dates: bool = True,
                 ref_option: bool = True,
                 extra_args: object = ()) -> object:
    """

    :param dates:
    :param ref_option:
    :param extra_args:
    :return:
    """

    ap = argparse.ArgumentParser()
    ap.add_argument("name", type=str)
    ap.add_argument("hemisphere", choices=("north", "south"))

    if dates:
        add_date_args(ap)

        # FIXME#11: not allowing this option currently
        # ap.add_argument("-d", "--date-ratio", type=float, default=1.0)

    ap.add_argument("-l", "--lag", type=int, default=2)
    ap.add_argument("-f", "--forecast", type=int, default=93)

    ap.add_argument("--abs",
                    help="Comma separated list of abs vars",
                    type=csv_arg,
                    default=[])
    ap.add_argument("--anom",
                    help="Comma separated list of abs vars",
                    type=csv_arg,
                    default=[])
    ap.add_argument("--trends",
                    help="Comma separated list of abs vars",
                    type=csv_arg,
                    default=[])
    ap.add_argument("--trend-lead",
                    help="Time steps in the future for linear trends",
                    type=int_or_list_arg,
                    default=93)

    for arg in extra_args:
        ap.add_argument(*arg[0], **arg[1])

    if ref_option:
        ap.add_argument("-r", "--ref",
                        help="Reference loader for normalisations etc",
                        default=None, type=str)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    ap.add_argument("-u", "--update-key",
                    default=None,
                    help="Add update key to processor to avoid overwriting default"
                         "entries in the loader configuration",
                    type=str)

    args = ap.parse_args()
    return args


def add_date_args(arg_parser: object):
    """

    :param arg_parser:
    """
    arg_parser.add_argument("-ns", "--train_start",
                            type=dates_arg, required=False, default=[])
    arg_parser.add_argument("-ne", "--train_end",
                            type=dates_arg, required=False, default=[])
    arg_parser.add_argument("-vs", "--val_start",
                            type=dates_arg, required=False, default=[])
    arg_parser.add_argument("-ve", "--val_end",
                            type=dates_arg, required=False, default=[])
    arg_parser.add_argument("-ts", "--test-start",
                            type=dates_arg, required=False, default=[])
    arg_parser.add_argument("-te", "--test-end", dest="test_end",
                            type=dates_arg, required=False, default=[])


def process_date_args(args: object) -> dict:
    """

    :param args:
    :return:
    """
    dates = dict(train=[], val=[], test=[])

    for dataset in ("train", "val", "test"):
        dataset_dates = collections.deque()

        for i, period_start in \
                enumerate(getattr(args, "{}_start".format(dataset))):
            period_end = getattr(args, "{}_end".format(dataset))[i]
            dataset_dates += [pd.to_datetime(date).date() for date in
                              pd.date_range(period_start,
                                            period_end, freq="D")]
        logging.info("Got {} dates for {}".format(len(dataset_dates),
                                                  dataset))

        dates[dataset] = sorted(list(dataset_dates))
    return dates




import inspect

from icenet.data.loaders.base import IceNetBaseDataLoader

import icenet.data.loaders.dask
import icenet.data.loaders.stdlib


class IceNetDataLoaderFactory:
    """

    """
    def __init__(self):
        self._loader_map = dict(
            dask=icenet.data.loaders.dask.DaskMultiWorkerLoader,
            dask_shared=icenet.data.loaders.dask.DaskMultiSharingWorkerLoader,
            standard=icenet.data.loaders.stdlib.IceNetDataLoader,
        )

    def add_data_loader(self, loader_name: str, loader_impl: object):
        """

        :param loader_name:
        :param loader_impl:
        """
        if loader_name not in self._loader_map:
            if IceNetBaseDataLoader in inspect.getmro(loader_impl):
                self._loader_map[loader_name] = loader_impl
            else:
                raise RuntimeError("{} is not descended from "
                                   "IceNetBaseDataLoader".
                                   format(loader_impl.__name__))
        else:
            raise RuntimeError("Cannot add {} as already in loader map".
                               format(loader_name))

    def create_data_loader(self, loader_name, *args, **kwargs):
        """

        :param loader_name:
        :param args:
        :param kwargs:
        :return:
        """
        return self._loader_map[loader_name](*args, **kwargs)

    @property
    def loader_map(self):
        return self._loader_map



import argparse

from icenet.data.loaders import IceNetDataLoaderFactory
from icenet.data.cli import add_date_args, process_date_args
from icenet.utils import setup_logging

"""

"""


@setup_logging
def create_get_args():
    implementations = list(IceNetDataLoaderFactory().loader_map)

    ap = argparse.ArgumentParser()
    ap.add_argument("name", type=str)
    ap.add_argument("hemisphere", choices=("north", "south"))

    ap.add_argument("-c", "--cfg-only", help="Do not generate data, "
                                             "only config", default=False,
                    action="store_true", dest="cfg")
    ap.add_argument("-d", "--dry",
                    help="Don't output files, just generate data",
                    default=False, action="store_true")
    ap.add_argument("-dt", "--dask-timeouts", type=int, default=120)
    ap.add_argument("-dp", "--dask-port", type=int, default=8888)
    ap.add_argument("-f", "--futures-per-worker", type=float, default=2.,
                    dest="futures")
    ap.add_argument("-fn", "--forecast-name", dest="forecast_name",
                    default=None, type=str)
    ap.add_argument("-fd", "--forecast-days", dest="forecast_days",
                    default=93, type=int)

    ap.add_argument("-i", "--implementation", type=str,
                    choices=implementations,
                    default=implementations[0])
    ap.add_argument("-l", "--lag", type=int, default=2)

    ap.add_argument("-ob", "--output-batch-size", dest="batch_size", type=int,
                    default=8)

    ap.add_argument("-p", "--pickup", help="Skip existing tfrecords",
                    default=False, action="store_true")
    ap.add_argument("-t", "--tmp-dir", help="Temporary directory",
                    default="/local/tmp", dest="tmp_dir", type=str)

    ap.add_argument("-v", "--verbose", action="store_true", default=False)
    ap.add_argument("-w", "--workers", help="Number of workers to use "
                                            "generating sets",
                    type=int, default=2)

    add_date_args(ap)
    args = ap.parse_args()
    return args


def create():
    args = create_get_args()
    dates = process_date_args(args)

    dl = IceNetDataLoaderFactory().create_data_loader(
        args.implementation,
        "loader.{}.json".format(args.name),
        args.forecast_name if args.forecast_name else args.name,
        args.lag,
        dates_override=dates
        if sum([len(v) for v in dates.values()]) > 0 else None,
        dry=args.dry,
        n_forecast_days=args.forecast_days,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        output_batch_size=args.batch_size,
        pickup=args.pickup,
        generate_workers=args.workers,
        dask_port=args.dask_port,
        futures_per_worker=args.futures)

    if args.cfg:
        dl.write_dataset_config_only()
    else:
        dl.generate()


def get_sample():
    raise NotImplementedError("get_sample is yet to be implemented")



import datetime
import logging
import os
import sys

from itertools import product

import ecmwfapi
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.interfaces.downloader import ClimateDownloader
from icenet.data.interfaces.utils import batch_requested_dates

"""

"""


class HRESDownloader(ClimateDownloader):
    """Climate downloader to provide CMIP6 reanalysis data from ESGF APIs

    :param identifier: how to identify this dataset

    """

    PARAM_TABLE = 128

    # Background on the use of forecast and observational data
    # https://confluence.ecmwf.int/pages/viewpage.action?pageId=85402030
    # https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation#ERA5:datadocumentation-Dateandtimespecification
    HRES_PARAMS = {
        "siconca":      (31, "siconc"), # sea_ice_area_fraction
        "tos":          (34, "sst"),    # sea surface temperature (actually
                                        # sst?)
        "zg":           (129, "z"),     # geopotential
        "ta":           (130, "t"),     # air_temperature (t)
        "hus":          (133, "q"),     # specific_humidity
        "psl":          (134, "sp"),    # surface_pressure
        "uas":          (165, "u10"),   # 10m_u_component_of_wind
        "vas":          (166, "v10"),   # 10m_v_component_of_wind
        "tas":          (167, "t2m"),   # 2m_temperature (t2m)
        # https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation#ERA5:datadocumentation-Meanrates/fluxesandaccumulations
        # https://apps.ecmwf.int/codes/grib/param-db/?id=175
        # https://confluence.ecmwf.int/pages/viewpage.action?pageId=197702790
        #
        # Mean rate/flux parameters in ERA5 (e.g. Table 4 for surface and
        # single levels) provide similar information to accumulations (e.g.
        # Table 3 for surface and single levels), except they are expressed as
        # temporal means, over the same processing periods, and so have units
        # of "per second".
        "rlds":         (175, "strd"),
        "rsds":         (169, "ssrd"),

        # plev  129.128 / 130.128 / 133.128
        # sfc   31.128 / 34.128 / 134.128 /
        #       165.128 / 166.128 / 167.128 / 169.128 / 177.128

        # ORAS5 variables in param-db (need to consider depth)
        #"thetao":       (151129, "thetao"),
        #"so":           (151130, "so"),
        # Better matches than equivalent X / Y parameters in param-db
        #"uo":           (151131, "uo"),
        #"vo":           (151132, "vo"),
    }

    # https://confluence.ecmwf.int/display/UDOC/Keywords+in+MARS+and+Dissemination+requests
    MARS_TEMPLATE = """
retrieve,
  class=od,
  date={date},
  expver=1,
  levtype={levtype},
  {levlist}param={params},
  step={step},
  stream=oper,
  time=12:00:00,
  type=fc,
  area={area},
  grid=0.25/0.25,
  target="{target}",
  format=netcdf
    """

    def __init__(self,
                 *args,
                 identifier: str = "mars.hres",
                 **kwargs):
        super().__init__(*args,
                         identifier=identifier,
                         **kwargs)

        self._server = ecmwfapi.ECMWFService("mars")

    def _single_download(self,
                         var_names: object,
                         pressures: object,
                         req_dates: object):
        """

        :param var_names:
        :param pressures:
        :param req_dates:
        :return:
        """

        for dt in req_dates:
            assert dt.year == req_dates[0].year

        downloads = []
        levtype = "plev" if pressures else "sfc"

        for req_batch in batch_requested_dates(req_dates, attribute="month"):
            req_batch = sorted(req_batch)
            request_month = req_batch[0].strftime("%Y%m")

            logging.info("Downloading month file {}".format(request_month))

            if req_batch[-1] - datetime.datetime.utcnow().date() == \
                    datetime.timedelta(days=-1):
                logging.warning("Not allowing partial requests at present, "
                                "removing {}".format(req_batch[-1]))
                req_batch = req_batch[:-1]

            request_target = os.path.join(
                self.base_path,
                self.hemisphere_str[0],
                "{}.{}.nc".format(levtype, request_month))

            os.makedirs(os.path.dirname(request_target), exist_ok=True)

            request = self.mars_template.format(
                area="/".join([str(s) for s in self.hemisphere_loc]),
                date="/".join([el.strftime("%Y%m%d") for el in req_batch]),
                levtype=levtype,
                levlist="levelist={},\n  ".format(pressures) if pressures else "",
                params="/".join(
                    ["{}.{}".format(
                        self.params[v][0],
                        self.param_table)
                     for v in var_names]),
                target=request_target,
                # We are only allowed date prior to -24 hours ago, dynamically
                # retrieve if date is today
                # TODO: too big - step="/".join([str(i) for i in range(24)]),
                step=0,
            )

            if not os.path.exists(request_target):
                logging.debug("MARS REQUEST: \n{}\n".format(request))

                try:
                    self._server.execute(request, request_target)
                except ecmwfapi.api.APIException:
                    logging.exception("Could not complete ECMWF request: {}")
                else:
                    downloads.append(request_target)
            else:
                logging.debug("Already have {}".format(request_target))
                downloads.append(request_target)

        logging.debug("Files downloaded: {}".format(downloads))

        ds = xr.open_mfdataset(downloads)
        ds = ds.resample(time='1D').mean()

        for var_name, pressure in product(var_names, pressures.split('/')
                                          if pressures else [None]):
            var = var_name if not pressure else \
                "{}{}".format(var_name, pressure)

            da = getattr(ds,
                         self.params[var_name][1])

            if pressure:
                da = da.sel(level=int(pressure))

            self.save_temporal_files(var, da)

        ds.close()

        if self.delete:
            for downloaded_file in downloads:
                if os.path.exists(downloaded_file):
                    logging.info("Removing {}".format(downloaded_file))
                    os.unlink(downloaded_file)

    def download(self):
        """

        """
        logging.info("Building request(s), downloading and daily averaging "
                     "from {} API".format(self.identifier.upper()))

        sfc_vars = [var for idx, var in enumerate(self.var_names)
                    if not self.pressure_levels[idx]]
        plev_vars = [var for idx, var in enumerate(self.var_names)
                     if self.pressure_levels[idx]]
        pressures = "/".join([str(s) for s in sorted(set(
            [p for ps in self.pressure_levels if ps for p in ps]))])

        # req_dates = self.filter_dates_on_data()

        dates_per_request = \
            batch_requested_dates(self._dates,
                                  attribute=self.group_dates_by)

        for req_batch in dates_per_request:
            if len(sfc_vars) > 0:
                self._single_download(sfc_vars, None, req_batch)

            if len(plev_vars) > 0:
                self._single_download(plev_vars, pressures, req_batch)

        logging.info("{} daily files downloaded".
                     format(len(self._files_downloaded)))

    def additional_regrid_processing(self,
                                     datafile: str,
                                     cube_ease: object):
        """

        :param datafile:
        :param cube_ease:
        """
        (datafile_path, datafile_name) = os.path.split(datafile)
        var_name = datafile_path.split(os.sep)[self._var_name_idx]

        if var_name == 'tos':
            # Overwrite maksed values with zeros
            logging.debug("MARS additional regrid: {}".format(var_name))
            cube_ease.data[cube_ease.data.mask] = 0.
            cube_ease.data[:, self._masks.get_land_mask()] = 0.
            cube_ease.data = cube_ease.data.data

        if var_name in ['rlds', 'rsds']:
            # FIXME: We're taking the mean across the hourly samples for the
            #  day in fc which needs to be comparative with the analysis product
            #  from ERA5. My interpretation is that this should be /24, but of
            #  course it doesn't work like that thanks to orbital rotation.
            #  We need to verify the exact mechanism for converting forecast
            #  values to reanalysis equivalents, but this rudimentary divisor
            #  should work in the meantime
            #
            #  FIXME FIXME FIXME
            cube_ease /= 12.

        if var_name.startswith("zg"):
            # https://apps.ecmwf.int/codes/grib/param-db/?id=129
            #
            # We want the geopotential height as per ERA5
            cube_ease /= 9.80665

    @property
    def mars_template(self):
        return getattr(self, "MARS_TEMPLATE")

    @property
    def params(self):
        return getattr(self, "HRES_PARAMS")

    @property
    def param_table(self):
        return getattr(self, "PARAM_TABLE")


class SEASDownloader(HRESDownloader):
    # TODO: step should be configurable for this downloader
    # TODO: unsure why, but multiple dates break the download with
    #  ERROR 89 (MARS_EXPECTED_FIELDS): Expected 4700, got 2350
    MARS_TEMPLATE = """
retrieve,
    class=od,
    date={date},
    expver=1,
    levtype={levtype},
    method=1,
    number=0/1/2/3/4/5/6/7/8/9/10/11/12/13/14/15/16/17/18/19/20/21/22/23/24,
    origin=ecmf,
    {levlist}param={params},
    step=0/to/2232/by/24,
    stream=mmsf,
    system=5,
    time=00:00:00,
    type=fc,
    target="{target}",
    format=netcdf,
    grid=0.25/0.25,
    area={area}"""

    def _single_download(self,
                         var_names: object,
                         pressures: object,
                         req_dates: object):
        """

        :param var_names:
        :param pressures:
        :param req_dates:
        :return:
        """

        for dt in req_dates:
            assert dt.year == req_dates[0].year

        downloads = []
        levtype = "plev" if pressures else "sfc"

        for req_date in req_dates:
            request_day = req_date.strftime("%Y%m%d")

            logging.info("Downloading daily file {}".format(request_day))

            request_target = os.path.join(
                self.base_path,
                self.hemisphere_str[0],
                "{}.{}.nc".format(levtype, request_day))
            os.makedirs(os.path.dirname(request_target), exist_ok=True)

            request = self.mars_template.format(
                area="/".join([str(s) for s in self.hemisphere_loc]),
                date=req_date.strftime("%Y-%m-%d"),
                levtype=levtype,
                levlist="levelist={},\n  ".format(pressures) if pressures else "",
                params="/".join(
                    ["{}.{}".format(
                        self.params[v][0],
                        self.param_table)
                     for v in var_names]),
                target=request_target,
            )

            if not os.path.exists(request_target):
                logging.debug("MARS REQUEST: \n{}\n".format(request))

                try:
                    self._server.execute(request, request_target)
                except ecmwfapi.api.APIException:
                    logging.exception("Could not complete ECMWF request: {}")
                else:
                    downloads.append(request_target)
            else:
                logging.debug("Already have {}".format(request_target))
                downloads.append(request_target)

        logging.debug("Files downloaded: {}".format(downloads))

        for download_filename in downloads:
            logging.info("Processing {}".format(download_filename))
            ds = xr.open_dataset(download_filename)
            ds = ds.mean("number")

            for var_name, pressure in product(var_names, pressures.split('/')
                                              if pressures else [None]):
                var = var_name if not pressure else \
                    "{}{}".format(var_name, pressure)

                da = getattr(ds, self.params[var_name][1])

                if pressure:
                    da = da.sel(level=int(pressure))

                self.save_temporal_files(var, da, date_format="%Y%m%d")

            ds.close()

        if self.delete:
            for downloaded_file in downloads:
                if os.path.exists(downloaded_file):
                    logging.info("Removing {}".format(downloaded_file))
                    os.unlink(downloaded_file)

    def save_temporal_files(self, var, da,
                            date_format=None,
                            freq=None):
        """

        :param var:
        :param da:
        :param date_format:
        :param freq:
        """
        var_folder = self.get_data_var_folder(var)

        req_date = pd.to_datetime(da.time.values[0])
        latlon_path, regridded_name = \
            self.get_req_filenames(var_folder,
                                   req_date,
                                   date_format=date_format)

        logging.info("Retrieving and saving {}".format(latlon_path))
        da.compute()
        da.to_netcdf(latlon_path)

        if not os.path.exists(regridded_name):
            self._files_downloaded.append(latlon_path)


def main(identifier, extra_kwargs=None):
    args = download_args()

    logging.info("ECMWF {} Data Downloading".format(identifier))
    cls = getattr(sys.modules[__name__], "{}Downloader".format(identifier))

    if extra_kwargs is None:
        extra_kwargs = dict()

    instance = cls(
        identifier="mars.{}".format(identifier.lower()),
        var_names=args.vars,
        pressure_levels=args.levels,
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        **extra_kwargs
    )
    instance.download()
    instance.regrid()
    instance.rotate_wind_data()


def seas_main():
    main("SEAS", dict(
        group_dates_by="day",
    ))


def hres_main():
    main("HRES")



import argparse
import collections
import glob
import logging
import os

import pandas as pd
import xarray as xr

from icenet.utils import setup_logging


def batch_requested_dates(dates: object,
                          attribute: str = "month") -> object:
    """

    TODO: should be using Pandas DatetimeIndexes / Periods for this, but the
     need to refactor slightly, and this is working for the moment

    :param dates:
    :param attribute:
    :return:
    """
    dates = collections.deque(sorted(dates))

    batched_dates = []
    batch = []

    while len(dates):
        if not len(batch):
            batch.append(dates.popleft())
        else:
            if getattr(batch[-1], attribute) == getattr(dates[0], attribute):
                batch.append(dates.popleft())
            else:
                batched_dates.append(batch)
                batch = []

    if len(batch):
        batched_dates.append(batch)

    if len(dates) > 0:
        raise RuntimeError("Batching didn't work!")

    return batched_dates


def reprocess_monthlies(source: str,
                        hemisphere: str,
                        identifier: str,
                        output_base: str,
                        dry: bool = False,
                        var_names: object = None):
    """

    :param source:
    :param hemisphere:
    :param identifier:
    :param output_base:
    :param dry:
    :param var_names:
    """
    if not var_names:
        var_names = []

    for var_name in var_names:
        var_path = os.path.join(source, hemisphere, var_name)
        files = glob.glob("{}/{}_*.nc".format(var_path, var_name))

        for file in files:
            _, year = os.path.basename(os.path.splitext(file)[0]).\
                          split("_")[0:2]

            try:
                year = int(year)

                if not (1900 < year < 2200):
                    logging.warning("File is not between 1900-2200, probably "
                                    "not something to convert: {}".format(year))
            except ValueError:
                logging.warning("Cannot derive year from {}".format(year))
                continue

            destination = os.path.join(output_base,
                                       identifier,
                                       hemisphere,
                                       var_name,
                                       str(year))

            if not os.path.exists(destination):
                os.makedirs(destination, exist_ok=True)

            logging.info("Processing {} from {} to {}".
                         format(var_name, year, destination))

            ds = xr.open_dataset(file)

            var_names = [name for name in list(ds.data_vars.keys())
                         if not name.startswith("lambert_")]

            var_names = set(var_names)
            logging.debug("Files have var names {} which will be renamed to {}".
                          format(", ".join(var_names), var_name))

            ds = ds.rename({k: var_name for k in var_names})
            da = getattr(ds, var_name)

            for date in da.time.values:
                date = pd.Timestamp(date)
                fname = '{:04d}_{:02d}_{:02d}.nc'. \
                    format(date.year, date.month, date.day)
                daily = da.sel(time=slice(date, date))

                output_path = os.path.join(destination, fname)

                if dry or os.path.exists(output_path):
                    continue
                else:
                    daily.to_netcdf(output_path)


def add_time_dim(source: str,
                 hemisphere: str,
                 identifier: str,
                 dry: bool = False,
                 var_names: object = []):
    """

    :param source:
    :param hemisphere:
    :param identifier:
    :param dry:
    :param var_names:
    """
    files = {}

    for var_name in var_names:
        var_path = os.path.join(source, identifier, hemisphere, var_name)

        if var_name not in files:
            files[var_name] = {}

        file_list = glob.glob("{}/*/*.nc".format(var_path))

        for path, filename in [os.path.split(el) for el in file_list]:
            if filename.startswith("{}_".format(var_name)):
                raise RuntimeError("{} starts with var name, we only want "
                                   "correctly named files to convert".
                                   format(filename))
            year = str(path.split(os.sep)[-1])

            if year not in files[var_name]:
                files[var_name][year] = []

            src = os.path.join(path, filename)
            dest = os.path.join(path, "{}_{}".format(var_name, filename))

            if not dry:
                try:
                    os.rename(src, dest)
                except OSError as e:
                    logging.exception("Not able to move file to temporary"
                                      "destination {}".format(dest))
                    raise e
            else:
                logging.info("{} -> {}".format(src, dest))

            files[var_name][year].append(dest)

    for year_files in [files[var][el] for var in files for el in files[var]]:
        if not dry:
            ds = xr.open_mfdataset(year_files,
                                   combine="nested",
                                   concat_dim="time",
                                   parallel=True)

            if "siconca" in year_files[0]:
                ds = ds.rename_vars({"siconca": "ice_conc"})
                ds = ds.sortby("time")
                ds['time'] = [pd.Timestamp(el) for el in
                              ds.indexes['time'].normalize()]

            for d in ds.time.values:
                dt = pd.to_datetime(d)
                date_str = dt.strftime("%Y_%m_%d")
                fpath = os.path.join(os.path.split(year_files[0])[0],
                                     "{}.nc".format(date_str))

                if not os.path.exists(fpath):
                    dw = ds.sel(time=slice(dt, dt))

                    logging.info("Writing {}".format(fpath))
                    dw.to_netcdf(fpath)
                else:
                    raise RuntimeError("Already exists: {}".format(fpath))

            ds.close()

            for orig_file in year_files:
                logging.info("Removing {}".format(orig_file))
                os.unlink(orig_file)
        else:
            logging.info("Would process out: {}".format(year_files))


@setup_logging
def get_args():
    """

    :return:
    """
    a = argparse.ArgumentParser()
    a.add_argument("-d", "--dry", default=False, action="store_true")
    a.add_argument("-o", "--output", default="./data")
    a.add_argument("-v", "--verbose", default=False, action="store_true")
    a.add_argument("source")
    a.add_argument("hemisphere", choices=["nh", "sh"])
    a.add_argument("identifier")
    a.add_argument("vars", nargs='+')
    return a.parse_args()


def add_time_dim_main():
    """CLI stub to sort out missing time dimensions in daily data
    """

    args = get_args()
    logging.info("Temporary solution for sorting missing time dim in files")

    if args.output != "./data":
        raise RuntimeError("output is not used for this command: {}".format(
            args.output))

    add_time_dim(args.source, args.hemisphere, args.identifier,
                 dry=args.dry,
                 var_names=args.vars)


def reprocess_main():
    """CLI stub solution for reprocessing monthly files
    """
    args = get_args()
    logging.info("Temporary solution for reprocessing monthly files")
    reprocess_monthlies(args.source, args.hemisphere, args.identifier,
                        output_base=args.output,
                        dry=args.dry,
                        var_names=args.vars)




import datetime as dt
import glob
import logging
import os

import pandas as pd
import xarray as xr

from icenet.process.forecasts import broadcast_forecast
from icenet.data.sic.mask import Masks


def get_seas_forecast_da(hemisphere: str,
                         date: str,
                         bias_correct: bool = True,
                         source_path: object = os.path.join(".", "data", "mars.seas"),
                         ) -> tuple:
    """
    Atmospheric model Ensemble 15-day forecast (Set III - ENS)

    :param hemisphere:
    :param date:
    :param bias_correct:
    :param source_path:
    """

    # TODO: why aren't we using SEASDownloader?
    # TODO: we could download here potentially

    seas_file = os.path.join(
        source_path,
        hemisphere,
        "siconca",
        "{}.nc".format(date.strftime("%Y%m%d")))
    seas_ds = xr.open_dataset(seas_file)

    return seas_ds.siconc


def get_forecast_ds(forecast_file: object,
                    forecast_date: str,
                    stddev: bool = False
                    ) -> tuple:
    """

    :param forecast_file:
    :param forecast_date:
    :param stddev:
    :returns tuple(fc_ds, obs_ds, land_mask):
    """
    forecast_date = pd.to_datetime(forecast_date)

    forecast_ds = xr.open_dataset(forecast_file)
    get_key = "sic_mean" if not stddev else "sic_stddev"

    forecast_ds = getattr(
        forecast_ds.sel(time=slice(forecast_date, forecast_date)),
        get_key)

    return forecast_ds


def filter_ds_by_obs(ds: object,
                     obs_da: object,
                     forecast_date: str) -> object:
    """

    :param ds:
    :param obs_da:
    :param forecast_date:
    :return:
    """
    forecast_date = pd.to_datetime(forecast_date)
    (start_date, end_date) = (
            forecast_date + dt.timedelta(days=int(ds.leadtime.min())),
            forecast_date + dt.timedelta(days=int(ds.leadtime.max()))
    )

    if len(obs_da.time) < len(ds.leadtime):
        logging.warning("Observational data not available for full range of "
                        "forecast leadtimes: {}-{} vs {}-{}".format(
                         obs_da.time.to_series()[0].strftime("%D"),
                         obs_da.time.to_series()[-1].strftime("%D"),
                         start_date.strftime("%D"),
                         end_date.strftime("%D")))
        (start_date, end_date) = (
            obs_da.time.to_series()[0],
            obs_da.time.to_series()[-1]
        )

    # We broadcast to get a nicely compatible dataset for plotting
    return broadcast_forecast(start_date=start_date,
                              end_date=end_date,
                              dataset=ds)


def get_obs_da(hemisphere: str,
               start_date: str,
               end_date: str,
               obs_source: object =
               os.path.join(".", "data", "osisaf"),
               ) -> object:
    """

    :param hemisphere:
    :param start_date:
    :param end_date:
    :param obs_source:
    :return:
    """
    obs_years = pd.Series(pd.date_range(start_date, end_date)).dt.year.unique()
    obs_dfs = [el for yr in obs_years for el in
               glob.glob(os.path.join(obs_source,
                                      hemisphere,
                                      "siconca", "{}.nc".format(yr)))]

    if len(obs_dfs) < len(obs_years):
        logging.warning("Cannot find necessary obs source files for {} in {}".
                        format(obs_years, obs_source))

    obs_ds = xr.open_mfdataset(obs_dfs)
    obs_ds = obs_ds.sel(time=slice(start_date, end_date))

    return obs_ds.ice_conc



import logging
import numpy as np
import matplotlib.pyplot as plt

################################################################################
# LEARNING RATE
################################################################################


def make_exp_decay_lr_schedule(rate: float,
                               start_epoch: int = 1,
                               end_epoch: object = np.inf) -> object:
    """ Returns an exponential learning rate function that multiplies by
    exp(-rate) each epoch after `start_epoch`.
    :param rate:
    :param start_epoch:
    :param end_epoch:
    :return:
    """

    def lr_scheduler_exp_decay(epoch, lr):
        """ Learning rate scheduler for fine tuning.
        Exponential decrease after start_epoch until end_epoch.
        :param epoch:
        :param lr:
        :return:
        """

        if start_epoch < epoch < end_epoch:
            lr = lr * np.math.exp(-rate)

        logging.info('\nSetting learning rate to: {}\n'.format(lr))

        return lr

    return lr_scheduler_exp_decay


###############################################################################
# PLOTTING
###############################################################################


def compute_heatmap(results_df: object,
                    model: object,
                    seed: object = 'NA',
                    metric: object = 'Binary accuracy') -> object:
    """
    Returns a binary accuracy heatmap of lead time vs. calendar month
    for a given model.
    :param results_df:
    :param model:
    :param seed:
    :param metric:
    :return:
    """

    month_names = np.array(['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                            'Jul', 'Aug', 'Sept', 'Oct', 'Nov', 'Dec'])

    # Mean over calendar month
    mean_df = results_df.loc[model, seed].reset_index().\
        groupby(['Calendar month', 'Leadtime']).mean()

    # Pivot
    heatmap_df = mean_df.reset_index().\
        pivot('Calendar month', 'Leadtime', metric).reindex(month_names)

    return heatmap_df


def arr_to_ice_edge_arr(arr: object,
                        thresh: object,
                        land_mask: object,
                        region_mask: object) -> object:

    """
    Compute a boolean mask with True over ice edge contour grid cells using
    matplotlib.pyplot.contour and an input threshold to define the ice edge
    (e.g. 0.15 for the 15% SIC ice edge or 0.5 for SIP forecasts). The contour
    along the coastline is removed using the region mask.
    :param arr:
    :param thresh:
    :param land_mask:
    :param region_mask:
    :return:
    """

    X, Y = np.meshgrid(np.arange(arr.shape[0]), np.arange(arr.shape[1]))
    X = X.T
    Y = Y.T

    cs = plt.contour(X, Y, arr, [thresh], alpha=0)  # Do not plot on any axes
    x = []
    y = []
    for p in cs.collections[0].get_paths():
        x_i, y_i = p.vertices.T
        x.extend(np.round(x_i))
        y.extend(np.round(y_i))
    x = np.array(x, int)
    y = np.array(y, int)
    ice_edge_arr = np.zeros(arr.shape, dtype=bool)
    ice_edge_arr[x, y] = True
    # Mask out ice edge contour that hugs the coastline
    ice_edge_arr[land_mask] = False
    ice_edge_arr[region_mask == 13] = False

    return ice_edge_arr


def arr_to_ice_edge_rgba_arr(arr: object,
                             thresh: object,
                             land_mask: object,
                             region_mask: object,
                             rgb: object) -> object:
    """

    :param arr:
    :param thresh:
    :param land_mask:
    :param region_mask:
    :param rgb:
    :return:
    """
    ice_edge_arr = arr_to_ice_edge_arr(arr, thresh, land_mask, region_mask)

    # Contour pixels -> alpha=1, alpha=0 elsewhere
    ice_edge_rgba_arr = np.zeros((*arr.shape, 4))
    ice_edge_rgba_arr[:, :, 3] = ice_edge_arr
    ice_edge_rgba_arr[:, :, :3] = rgb

    return ice_edge_rgba_arr
